from pathlib import Path
from typing import Dict, List

import pytest
import torch
import transformers

from tapeagents.finetune.data import collate, preprocess_fn

res_path = Path(__file__).parent.resolve() / "res"


@pytest.fixture
def tokenizer():
    return transformers.AutoTokenizer.from_pretrained(res_path / "tokenizer/starcoder/")


def test_preprocess_text(tokenizer):
    short_string = "hello world"
    sample = preprocess_fn({"text": short_string}, tokenizer, seq_length=2048)
    assert sample == {
        "input_ids": [7656, 5788],
        "attention_mask": [1, 1],
        "offset_mapping": [(0, 5), (5, 11)],
        "labels": [7656, 5788],
    }


def test_preprocess_spans(tokenizer):
    short_string = "hello world over here"
    sample = preprocess_fn({"text": short_string, "predicted_spans": [[6, 11], [18, 21]]}, tokenizer, seq_length=2048)
    assert sample == {
        "input_ids": [7656, 5788, 2288, 2442],
        "attention_mask": [1, 1, 1, 1],
        "offset_mapping": [(0, 5), (5, 11), (11, 16), (16, 21)],
        "labels": [-100, 5788, -100, 2442],
    }


def test_preprocess_wrong_span(tokenizer):
    short_string = "hello world over here"
    with pytest.raises(ValueError):
        preprocess_fn({"text": short_string, "predicted_spans": [[6, 11], [18, 22]]}, tokenizer, seq_length=2048)


def test_preprocess_truncation(tokenizer):
    string = "hello world over here" * 15
    seq_length = 20
    sample = preprocess_fn({"text": string}, tokenizer, seq_length=seq_length)
    for k, v in sample.items():
        assert len(v) == seq_length, f"Truncation failed for field {k}"


def test_preprocess_long(tokenizer):
    long_string = "a = 1 + 2 * 3 - 4 / 5\n" * 500
    seq_length = 2048
    sample = preprocess_fn({"text": long_string}, tokenizer, seq_length=seq_length)
    for k, v in sample.items():
        assert len(v) == seq_length, f"Truncation failed for field {k}"


def get_samples(tokenizer, seq_length) -> List[Dict[str, List[int]]]:
    short_string = "hello world"
    mid_string = "hello world over here"
    long_string = "this is the test of the padding on the longest string"
    samples = [
        preprocess_fn({"text": text}, tokenizer, seq_length=seq_length)
        for text in [short_string, long_string, mid_string]
    ]
    for s in samples:
        s.pop("offset_mapping")
    return samples  # type: ignore


def test_collate_clm(tokenizer):
    samples = get_samples(tokenizer, seq_length=16)
    batch = collate(samples, tokenizer, pad_to_multiple_of=4)
    for v in batch.values():
        assert isinstance(v, torch.Tensor)
    batch = {k: v.tolist() for k, v in batch.items()}
    assert batch == {
        "input_ids": [
            [7656, 5788, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [578, 438, 322, 894, 432, 322, 3714, 544, 322, 30698, 802, 0],
            [7656, 5788, 2288, 2442, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
        "attention_mask": [
            [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
            [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
        "labels": [
            [7656, 5788, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100],
            [578, 438, 322, 894, 432, 322, 3714, 544, 322, 30698, 802, -100],
            [7656, 5788, 2288, 2442, -100, -100, -100, -100, -100, -100, -100, -100],
        ],
    }


def test_collate_left_padding(tokenizer):
    tokenizer.padding_side = "left"
    left_pad_samples = get_samples(tokenizer, seq_length=16)
    batch = collate(left_pad_samples, tokenizer)
    for v in batch.values():
        assert isinstance(v, torch.Tensor)
    batch = {k: v.tolist() for k, v in batch.items()}
    assert batch == {
        "input_ids": [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7656, 5788],
            [0, 0, 0, 0, 0, 578, 438, 322, 894, 432, 322, 3714, 544, 322, 30698, 802],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7656, 5788, 2288, 2442],
        ],
        "attention_mask": [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
            [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        ],
        "labels": [
            [-100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, 7656, 5788],
            [-100, -100, -100, -100, -100, 578, 438, 322, 894, 432, 322, 3714, 544, 322, 30698, 802],
            [-100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, 7656, 5788, 2288, 2442],
        ],
    }


def test_collate_multiple_of_16(tokenizer):
    non_16_samples = get_samples(tokenizer, seq_length=5)
    batch = collate(non_16_samples, tokenizer)
    for v in batch.values():
        assert isinstance(v, torch.Tensor)
    batch = {k: v.tolist() for k, v in batch.items()}
    assert batch == {
        "input_ids": [
            [7656, 5788, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [578, 438, 322, 894, 432, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [7656, 5788, 2288, 2442, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
        "attention_mask": [
            [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
        "labels": [
            [7656, 5788, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100],
            [578, 438, 322, 894, 432, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100],
            [7656, 5788, 2288, 2442, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100],
        ],
    }


if __name__ == "__main__":
    _tokenizer = transformers.AutoTokenizer.from_pretrained(res_path / "tokenizer/starcoder/")
    test_preprocess_text(_tokenizer)
    test_preprocess_spans(_tokenizer)
    test_preprocess_wrong_span(_tokenizer)
    test_preprocess_truncation(_tokenizer)
    test_preprocess_long(_tokenizer)
    test_collate_clm(_tokenizer)
    test_collate_left_padding(_tokenizer)
    test_collate_multiple_of_16(_tokenizer)



import tempfile
from pathlib import Path

from tapeagents.batch import generate_tapes
from tapeagents.dialog_tape import AssistantStep, DialogTape, SystemStep, UserStep
from tapeagents.environment import EmptyEnvironment
from tapeagents.llms import LLM, TrainableLLM

from .llama_agent import LLAMAChatBot
from .llama_user import LLAMAUserModel


def try_continue_tapes(llm: LLM):
    agent = LLAMAChatBot.create(llm)
    follow_up_user = LLAMAUserModel(
        instruction="come up with a follow-up message that a user may write to their conversational assistant",
        llms={"default": llm},
    )
    change_topic_user = LLAMAUserModel(
        instruction="change the topic of the conversation to something completely different",
        llms={"default": llm},
    )

    tape = DialogTape(
        context=None,
        steps=[
            SystemStep(content="You are a helpful assistant."),
            UserStep(content="Tell me about ServiceNow"),
            AssistantStep(content="It is a very successful enterprise software company"),
        ],
    )

    layers_conf = [
        [(follow_up_user, 3)],
        [(follow_up_user, 3), (change_topic_user, 3)],
    ]
    path = Path(tempfile.mktemp())
    generate_tapes(
        agent,
        [tape],
        EmptyEnvironment(),
        layers_conf,  # type: ignore
        path,
    )

    print(f"Go check the output at {path}")


if __name__ == "__main__":
    try_continue_tapes(
        TrainableLLM(
            base_url="https://api.together.xyz",
            model_name="meta-llama/Llama-3-8b-chat-hf",
            parameters=dict(temperature=0.7, max_tokens=512),
        )
    )



import json
import logging
import os
import sys
from pathlib import Path

import torch
from peft.mapping import get_peft_model
from peft.peft_model import PeftModel
from peft.tuners.lora import LoraConfig
from peft.utils.other import prepare_model_for_kbit_training
from peft.utils.save_and_load import set_peft_model_state_dict
from safetensors.torch import load_file
from transformers import AutoModelForCausalLM, AutoTokenizer

logger = logging.getLogger(__name__)


def has_lora_checkpoint(current_dir: Path) -> bool:
    return os.path.exists(current_dir / "adapter_model.bin") or os.path.exists(
        current_dir / "adapter_model.safetensors"
    )


def prepare_model_for_bf16_training(model, use_gradient_checkpointing=False, gradient_checkpointing_kwargs=None):
    logger.info("Prepare LoRA for BF16 training")
    for _, param in model.named_parameters():
        # freeze base model's layers
        param.requires_grad = False

    for name, param in model.named_parameters():
        # upcast LM head and layernorms
        if any([k in name for k in ["lm_head", "wte", "ln_"]]):
            param.data = param.data.to(torch.float32)

    if use_gradient_checkpointing:
        # For backward compatibility
        if hasattr(model, "enable_input_require_grads"):
            model.enable_input_require_grads()
        else:

            def make_inputs_require_grad(module, input, output):
                output.requires_grad_(True)

            model.get_input_embeddings().register_forward_hook(make_inputs_require_grad)
        model.gradient_checkpointing_enable()

    return model


def prepare_lora_model(lora_config, model, gradient_checkpointing) -> PeftModel:
    logger.info("Prepare LoRA adapters")
    lora_prepare_fn = (
        prepare_model_for_kbit_training
        if lora_config.base_model_8bit or lora_config.base_model_4bit
        else prepare_model_for_bf16_training
    )
    all_params = model.num_parameters()
    model = lora_prepare_fn(
        model, use_gradient_checkpointing=gradient_checkpointing, gradient_checkpointing_kwargs={"use_reentrant": True}
    )
    lora_config = LoraConfig(
        inference_mode=False,
        r=lora_config.r,
        lora_alpha=lora_config.alpha,
        lora_dropout=lora_config.dropout,
        bias=lora_config.bias,
        task_type=lora_config.task_type,
        target_modules=list(lora_config.target_modules),
    )
    model = get_peft_model(model, lora_config)
    trainable_params = model.num_parameters(only_trainable=True)
    logger.info(
        f"LoRA trainable params: {trainable_params:_} of {all_params:_} || trainable%: {100 * trainable_params / all_params:.2f}"
    )
    return model  # type: ignore


def is_lora_checkpoint(lora_model_path):
    lora_model_config = os.path.join(lora_model_path, "adapter_config.json")
    return os.path.exists(lora_model_config)


def get_base_model_name(lora_model_path):
    assert is_lora_checkpoint(lora_model_path)
    lora_model_config = os.path.join(lora_model_path, "adapter_config.json")
    with open(lora_model_config) as f:
        lora_config = json.load(f)
    return lora_config["base_model_name_or_path"]


def lora_load_and_merge(lora_model_path, **kwargs):
    base_model_name_or_path = get_base_model_name(lora_model_path)
    logger.info(f"Load base model {base_model_name_or_path}")
    base_model = AutoModelForCausalLM.from_pretrained(base_model_name_or_path, **kwargs)
    model = PeftModel.from_pretrained(base_model, lora_model_path)
    logger.info(f"Merge adapters from {lora_model_path}")
    return model.merge_and_unload()  # type: ignore


def lora_save(checkpoint_folder, model):
    model.save_pretrained(checkpoint_folder)
    logger.info(f"Saved LoRA model to {checkpoint_folder}")


def lora_load(checkpoint_folder, model):
    adapter_path = os.path.join(checkpoint_folder, "adapter_model.safetensors")
    if os.path.exists(adapter_path):
        adapters_weights = load_file(adapter_path)
    else:
        adapter_path = os.path.join(checkpoint_folder, "adapter_model.bin")
        adapters_weights = torch.load(adapter_path)
    set_peft_model_state_dict(model, adapters_weights)
    logger.info(f"Loaded LoRA model from {checkpoint_folder}.")
    return model


def apply_lora(model, lora_model_path):
    lora_model_config = os.path.join(lora_model_path, "adapter_config.json")
    assert os.path.exists(lora_model_config)
    with open(lora_model_config) as f:
        lora_config = json.load(f)
    scaling = lora_config["lora_alpha"] / lora_config["r"]
    lora_model_file = os.path.join(lora_model_path, "adapter_model.bin")
    assert os.path.exists(lora_model_file)
    lora_weights = torch.load(lora_model_file)
    for layer_name, layer_weights in model.named_parameters():
        lora_a = None
        lora_b = None
        for k, v in lora_weights.items():
            if layer_name in k:
                if "lora_A" in k:
                    lora_a = v
                elif "lora_B" in k:
                    lora_b = v
            if lora_a is not None and lora_b is not None:
                wdiff = (lora_b @ lora_a) * scaling
                layer_weights.data += wdiff
                break


def merge_lora(lora_model_path):
    if lora_model_path[-1] == "/":
        lora_model_path = lora_model_path[:-1]
    assert os.path.isdir(lora_model_path), f"{lora_model_path} is not a dir"
    lora_model_config = os.path.join(lora_model_path, "adapter_config.json")
    assert os.path.exists(lora_model_config), f"{lora_model_config} does not exists"

    logger.info(f"Merge lora checkpoint {lora_model_path}")
    model = lora_load_and_merge(lora_model_path, torch_dtype=torch.bfloat16, low_cpu_mem_usage=True)
    tokenizer = AutoTokenizer.from_pretrained(lora_model_path)

    tmp_dir = f"{lora_model_path}_merged"
    logger.info(f"Save merged model to {tmp_dir}")
    model.save_pretrained(tmp_dir, safe_serialization=True)
    tokenizer.save_pretrained(tmp_dir)

    os.rename(lora_model_path, f"{lora_model_path}_lora")
    os.rename(tmp_dir, lora_model_path)
    logger.info(f"Merged model saved to {lora_model_path}")


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Merging lora weights: python lora.py <lora_weights_dir>"
    merge_lora(sys.argv[1])



import json
import logging
from typing import Literal, Type

import numpy as np
import pydantic
from pydantic import BaseModel

from tapeagents.agent import Node, ObservationMaker, TapeMetadata
from tapeagents.core import Action, Error, MakeObservation, Prompt, SetNextNode, StepMetadata, Tape, Thought
from tapeagents.dialog_tape import UserStep
from tapeagents.llms import LLMStream

from .environment import FormFillerEnvironment
from .error import FormFillerStateError
from .state import compute_form_filler_state
from .steps import RequestFunctionParameters
from .tape import FormFillerTape
from .user_simulator_filters import (
    AlwaysTrue,
    extract_placeholder_keys,
    get_step_instruction_params,
)
from .utils import get_step_cls, render_chat_template

logger = logging.getLogger(__name__)


class SampleUserInstructionThought(Thought):
    kind: Literal["sample_user_instruction_thought"] = "sample_user_instruction_thought"  # type: ignore
    instruction_alias: str
    instruction: str


class UserSimulatorError(Action, Error):
    kind: Literal["user_simulator_error"] = "user_simulator_error"  # type: ignore
    error: str


# The user simulator tape wraps the formfiller tape as its context
# It produces (potentially some intermediate steps) and a final step
UserSimulatorTape = Tape[FormFillerTape, SampleUserInstructionThought | MakeObservation[UserStep] | UserSimulatorError]


class InstructionSampler(BaseModel):
    instructions: dict[str, str]
    weights: dict[str, float]

    @pydantic.model_validator(mode="after")
    def validate_weights(self):
        # Verify that weights are positive and keys match
        if not all(w >= 0 for w in self.weights.values()):
            raise ValueError("Weights must be non-negative.")
        if set(self.instructions.keys()) != set(self.weights.keys()):
            raise ValueError("Instruction keys and weights keys must match.")
        return self

    def sample_instruction(self) -> SampleUserInstructionThought:
        import random

        if not self.instructions or not self.weights:
            raise ValueError("Instructions and weights must be provided for sampling.")

        normalized_weights = [self.weights[key] for key in self.instructions.keys()]
        total_weight = sum(normalized_weights)
        if total_weight == 0:
            raise ValueError("Total weight must be greater than zero for sampling.")
        normalized_weights = [w / total_weight for w in normalized_weights]

        sampled_key = random.choices(list(self.instructions.keys()), weights=normalized_weights, k=1)[0]

        sample_user_instruction_thought = SampleUserInstructionThought(
            instruction_alias=sampled_key,
            instruction=self.instructions[sampled_key],
        )

        return sample_user_instruction_thought

    @classmethod
    def from_str(cls, instruction_str: str) -> "InstructionSampler":
        # When given a plain string, use it as a single instruction
        return cls(instructions={"default_instruction": instruction_str}, weights={"default_instruction": 1.0})


class UserSimulatorAgent(ObservationMaker[FormFillerTape, UserSimulatorTape]):
    nodes: list[Node] = []
    behavior_alias: str
    prompt_template: Prompt
    instruction: InstructionSampler
    step_types_chain: list[str] = []
    _step_types_chain: list[Type] = []  # same but the str are converted to types
    step_filters: list = []
    _step_filters: list = []  # same but padded to match the length of step_types_chain

    @classmethod
    def create(cls, instruction: str | InstructionSampler, **kwargs) -> "UserSimulatorAgent":
        if isinstance(instruction, str):
            instruction = InstructionSampler.from_str(instruction)

        nodes = [
            SampleUserInstructionNode(),
            UserSimulatorMainNode(),
        ]
        user_simulator_agent = super().create(name="user_simulator", nodes=nodes, instruction=instruction, **kwargs)
        user_simulator_agent._step_types_chain = [get_step_cls(name) for name in user_simulator_agent.step_types_chain]

        assert len(user_simulator_agent.step_filters) <= len(
            user_simulator_agent._step_types_chain
        ), "Step filters must be less than or equal to the step types chain"

        # Fill step_filters with True to match the length of step_types_chain
        user_simulator_agent._step_filters = user_simulator_agent.step_filters + [
            AlwaysTrue()
            for __ in range(len(user_simulator_agent._step_types_chain) - len(user_simulator_agent.step_filters))
        ]

        return user_simulator_agent

    def make_own_tape(self, tape: FormFillerTape) -> UserSimulatorTape:
        # Hacky stuff to remove things like setnextnode so that filters and placeholder work
        tape_without_setnextnode = tape.model_copy(
            update={"steps": [step for step in tape.steps if not isinstance(step, SetNextNode)]}
        )

        # This just wraps the input formfiller tape into a user simulator tape's context
        user_simulator_tape = UserSimulatorTape(
            metadata=TapeMetadata(
                parent_id=tape.metadata.id,
                author="user_simulator",
            ),
            context=tape_without_setnextnode,  # use input tape as context
            steps=[],
        )
        return user_simulator_tape

    def can_continue(self, tape: FormFillerTape) -> bool:
        # Hacky stuff to remove things like setnextnode so that filters and placeholder work
        tape_without_setnextnode = tape.model_copy(
            update={"steps": [step for step in tape.steps if not isinstance(step, SetNextNode)]}
        )

        # If the tape is too short, we can't continue
        if len(tape_without_setnextnode.steps) < len(self.step_types_chain):
            return False

        # Compute the state
        state = compute_form_filler_state(tape_without_setnextnode)
        if isinstance(state, FormFillerStateError):
            logger.error(
                f"Failed to compute state for tape_without_setnextnode {tape_without_setnextnode.metadata.id}: {state}"
            )
            return False

        # only assess the last steps
        last_steps = tape_without_setnextnode.steps[-len(self._step_types_chain) :]
        offset = len(tape_without_setnextnode.steps) - len(last_steps)

        # Verify step class[k]
        target_step_classes = [step_type.__name__ for step_type in self._step_types_chain]
        actual_step_classes = [step.__class__.__name__ for step in last_steps]

        # Verify step class[k] and step filter[k] match
        for i, (step, step_type, step_filter) in enumerate(zip(last_steps, self._step_types_chain, self._step_filters)):
            actual_index = i + offset
            if not isinstance(step, step_type):
                logger.debug(
                    f"Step {actual_index} of tape_without_setnextnode {tape_without_setnextnode.metadata.id} is not of type {step_type}: "
                )
                logger.debug(f"Expected: {target_step_classes}, got: {actual_step_classes}")
                return False
            if not step_filter.filter(state, step):
                logger.debug(
                    f"Step {actual_index} of tape_without_setnextnode {tape_without_setnextnode.metadata.id} failed the filter {repr(step_filter)}"
                )
                return False

        logger.debug(f"Tape {tape.metadata.id} is continuable")
        return True


class SampleUserInstructionNode(Node):
    name: str = "sample_instruction"

    # No prompt needed
    def generate_steps(self, agent: UserSimulatorAgent, tape: UserSimulatorTape, llm_stream: LLMStream):
        instruction_thought = agent.instruction.sample_instruction()
        yield instruction_thought


class UserSimulatorMainNode(Node):
    name: str = "main_node"

    def make_prompt(self, agent: UserSimulatorAgent, tape: UserSimulatorTape) -> Prompt:
        # Render the prompt template
        assert isinstance(tape.context, FormFillerTape), "UserSimulatorTape must have FormFillerTape context"
        assert isinstance(
            tape.steps[-1], SampleUserInstructionThought
        ), "Last step must be SampleUserInstructionThought"

        # Put the rendered dialogue and instruction into the main prompt template
        rendered_dialogue = tape.context.render_dialogue_as_text()
        instruction = tape.steps[-1].instruction
        prompt_template_with_instruction = render_chat_template(
            agent.prompt_template.messages,
            dict(
                text=rendered_dialogue,
                instruction=instruction,
            ),
        )

        # Now get the placeholders for the instruction
        placeholder_values = {}

        # 1) Get placeholders for function schemas
        # Use a hack to access the envrionemnt directly to get schemas
        assert tape.context.context is not None
        environment = FormFillerEnvironment.from_spec(tape.context.context.env_spec)
        sampled_schema = np.random.choice(list(environment.available_schemas))
        parameters = environment.available_schemas[sampled_schema].with_replaced_refs.model_dump()["parameters"]
        schema_placeholders = {
            "schema_description": environment.available_schemas[sampled_schema].description,
            "schema_parameters": json.dumps(parameters, indent=2),
            "schema_name": environment.available_schemas[sampled_schema].name,
            "all_schema_descriptions": json.dumps(
                dict((schema.name, schema.description) for schema in environment.available_schemas.values()), indent=2
            ),
            "all_schema_names": json.dumps(
                [schema.name for schema in environment.available_schemas.values()], indent=2
            ),
        }
        placeholder_values.update(schema_placeholders)

        # 2) Optionally, get placeholders on RequestFunctionParameters
        # In some cases, the user simulator assumes steps[-2] is RequestFunctionParameters
        # and uses it to generate extra placeholders
        if len(agent._step_types_chain) >= 2 and agent._step_types_chain[-2] == RequestFunctionParameters:
            state = compute_form_filler_state(tape.context)
            assert not isinstance(state, FormFillerStateError), "State should not be an error"

            request_function_parameters_step = tape.context.steps[-2]
            assert isinstance(
                request_function_parameters_step, RequestFunctionParameters
            ), f"Expected RequestFunctionParameters, got {request_function_parameters_step}"

            extra_placeholders = get_step_instruction_params(request_function_parameters_step, state)
            placeholder_values.update(extra_placeholders)

        # FINALLY: replace the placeholders in the prompt template
        rendered_prompt = render_chat_template(prompt_template_with_instruction, placeholder_values)

        remaining_placeholders = extract_placeholder_keys(rendered_prompt)
        assert not remaining_placeholders, f"Remaining placeholders in prompt: {remaining_placeholders}"

        logger.debug(f"Rendered prompt: {rendered_prompt}")
        return Prompt(messages=rendered_prompt)

    def generate_steps(self, agent: UserSimulatorAgent, tape: UserSimulatorTape, llm_stream: LLMStream):
        user_utterance = llm_stream.get_text()
        step = UserStep(
            content=user_utterance,
            metadata=StepMetadata(other={"alias": agent.behavior_alias}),
        )
        yield MakeObservation[UserStep](new_observation=step)



"""
Module contains the main loops of the agent-environment interaction and replay functions.
"""

import enum
import logging
from typing import Generator, Generic

from hydra.utils import instantiate
from omegaconf import DictConfig
from pydantic import BaseModel, Field
from termcolor import colored

from tapeagents.config import is_debug_mode

from .agent import Agent
from .core import AgentEvent, Observation, Step, StopStep, TapeType
from .environment import Environment, ExternalObservationNeeded, NoActionsToReactTo, ToolCollectionEnvironment
from .renderers import step_view
from .utils import FatalError, diff_dicts

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MainLoopStatus(enum.Enum):
    OK = "ok"
    FINISHED = "finished"
    NO_ACTIONS = "no_actions"
    EXTERNAL_INPUT_NEEDED = "external_input_needed"
    UNKNOWN_ACTION = "unknown_action"


class MainLoopEvent(BaseModel, Generic[TapeType]):
    # TODO: validate that event must have only either the agent or the environment data
    agent_event: AgentEvent[TapeType] | None = Field(default=None, description="Propagates all the agent events")
    agent_tape: TapeType | None = Field(default=None, description="The tape after agent.run()")
    observation: Observation | None = Field(
        default=None, description="One of the observations produced by environment.react()"
    )
    env_tape: TapeType | None = Field(default=None, description="The tape after environment.react()")
    status: MainLoopStatus = MainLoopStatus.OK


class MainLoopStream(Generic[TapeType]):
    def __init__(self, generator: Generator[MainLoopEvent[TapeType], None, None]):
        self.generator = generator

    def __bool__(self):
        return self.generator is not None

    def __iter__(self):
        if self.generator is None:
            raise ValueError("can't iterate a null stream")
        return self

    def __next__(self) -> MainLoopEvent:
        if self.generator is None:
            raise StopIteration
        return next(self.generator)

    def agent_events(self) -> Generator[AgentEvent[TapeType], None, None]:
        for event in self:
            if event.agent_event:
                yield event.agent_event

    def get_final_tape(self) -> TapeType:
        """Return the last tape by either the agent or the environment."""
        last_final_tape = None
        for event in self:
            if event.agent_tape:
                last_final_tape = event.agent_tape
            if event.env_tape:
                last_final_tape = event.env_tape
        if last_final_tape is not None:
            return last_final_tape
        raise ValueError("No tape by either the agent or the environment")


def get_agent_and_env_from_config(cfg: DictConfig) -> tuple[Agent, ToolCollectionEnvironment]:
    environment: ToolCollectionEnvironment = instantiate(cfg.environment)
    agent: Agent = instantiate(
        cfg.agent, known_actions=environment.actions(), tools_description=environment.tools_description()
    )
    return agent, environment


def main_loop(
    agent: Agent[TapeType],
    start_tape: TapeType,
    environment: Environment,
    max_loops: int = -1,
) -> MainLoopStream[TapeType]:
    """
    Main loop of the agent-environment interaction. The agent is run on the tape, then the environment reacts to the
    agent's tape, then the agent is run on the environment's tape, and so on.
    The loop stops when the agent emits a final step or the environment emits a final step,
    or the maximum number of loops is reached.

    :param agent: Agent object
    :param start_tape: initial tape
    :param environment: Environment object
    :param max_loops: maximum number of loops, -1 for infinite

    :return: generator of MainLoopEvent objects
    """

    def _implementation():
        if is_debug_mode():
            logger.setLevel(logging.DEBUG)
        n_loops = 0
        tape = start_tape
        event = None
        while n_loops < max_loops or max_loops == -1:
            # --- RUN THE AGENT ---
            for event in agent.run(tape):
                yield MainLoopEvent(agent_event=event)
                if event.step:
                    logger.info(colored(f"AGENT: {step_view(event.step)}", "green"))
                if event.final_tape:
                    break
            assert event and event.final_tape
            agent_tape = event.final_tape
            yield MainLoopEvent(agent_tape=agent_tape)

            # --- RUN THE ENVIRONMENT ---
            if any([isinstance(step, StopStep) for step in agent_tape.steps]):
                logger.info(f"Agent emitted final step {agent_tape.steps[-1]}")
                yield MainLoopEvent(status=MainLoopStatus.FINISHED)
                return
            try:
                tape = environment.react(agent_tape)
            except NoActionsToReactTo:
                yield MainLoopEvent(status=MainLoopStatus.NO_ACTIONS)
                return
            except ExternalObservationNeeded:
                yield MainLoopEvent(status=MainLoopStatus.EXTERNAL_INPUT_NEEDED)
                return
            for observation in tape[len(agent_tape) :]:
                logger.info(colored(f"ENV: {step_view(observation, trim=True)}", "yellow"))
                yield MainLoopEvent(observation=observation)
            yield MainLoopEvent[TapeType](env_tape=tape)

            # --- REPEAT ---
            n_loops += 1

    return MainLoopStream(_implementation())


def replay_tape(
    agent: Agent[TapeType],
    tape: TapeType,
    env: Environment[TapeType] | None = None,
    start_tape: TapeType | None = None,
    reuse_observations: bool = False,
    stop_on_mismatch: bool = True,
) -> bool:
    """
    Replay the tape with the agent and compare the steps with the old tape.
    Count mismatches and print diffs of them.

    :param agent: Agent object
    :param tape: Old tape object
    :param env: Environment object
    :param start_tape: initial tape, if None, the first observations of the tape are used
    :param reuse_observations: reuse observations from the tape instead of calling the environment
    :param stop_on_mismatch: stop the replay on the first mismatch

    :return: True if all steps match, False otherwise
    """
    if env is None and not reuse_observations:
        raise ValueError("Environment is required when not reusing observations")
    match: bool = True
    if start_tape is None:
        start_steps: list[Step] = []
        for step in tape.steps:
            if isinstance(step, Observation):
                start_steps.append(step)
            else:
                break
        start_tape = tape.model_copy(update=dict(steps=start_steps))

    event = None
    new_tape: TapeType = start_tape
    new_steps_count = len(start_tape)
    while new_steps_count < len(tape):
        for event in agent.run(new_tape):
            if event.step:
                step_dict = event.step.llm_dict()
                if new_steps_count >= len(tape.steps):
                    logger.error(f"Extra step {new_steps_count} from agent, kind: {step_dict.get('kind')}")
                    match = False
                    if stop_on_mismatch:
                        return False
                    break
                old_step_dict = tape.steps[new_steps_count].llm_dict()
                new_steps_count += 1
                kind = step_dict.get("kind")
                old_kind = old_step_dict.get("kind")
                if kind != old_kind:
                    logger.error(
                        f"Step {new_steps_count} kind mismatch: Old {old_kind}, New {kind}\nOld step: {old_step_dict}\nNew step: {step_dict}"
                    )
                    match = False
                    if stop_on_mismatch:
                        return False
                elif old_step_dict != step_dict:
                    logger.error(f"Step {new_steps_count} mismatch")
                    logger.error(diff_dicts(old_step_dict, step_dict))
                    match = False
                    if stop_on_mismatch:
                        return False
                else:
                    logger.debug(f"Step {new_steps_count} ok")
            if event.final_tape:
                break
        assert event and event.final_tape
        agent_tape = event.final_tape
        new_tape = agent_tape
        if isinstance(new_tape.steps[-1], StopStep):
            logger.info("Agent emitted final step, stop")
            break

        if reuse_observations:
            observations: list[Step] = []
            for step in tape.steps[new_steps_count:]:
                if isinstance(step, Observation):
                    observations.append(step)
                else:
                    break
            if len(observations):
                logger.debug(f"Reusing {len(observations)} observations from tape")
            new_tape = agent_tape + observations
            new_steps_count += len(observations)
        else:
            assert env is not None
            new_tape = env.react(agent_tape)
            observations = new_tape.steps[len(agent_tape) :]
            for observation in observations:
                step_dict = observation.llm_dict()
                old_step_dict = tape.steps[new_steps_count].llm_dict()
                new_steps_count += 1
                if old_step_dict != step_dict:
                    logger.error(f"Observation {new_steps_count} mismatch")
                    logger.error(diff_dicts(old_step_dict, step_dict))
                    match = False
                    if stop_on_mismatch:
                        return False
                else:
                    logger.debug(f"Observation {new_steps_count} ok")

                if isinstance(observation, StopStep):
                    logger.info(f"Environment emitted final step {observation}")
                    break
        if isinstance(new_tape.steps[-1], StopStep):
            logger.info("Env emitted final step, stop")
            break
    if new_steps_count != len(tape.steps):
        logger.error(f"New tape has {new_steps_count} steps, old tape has {len(tape.steps)}")
        match = False
    return match


def replay_tapes(
    agent: Agent[TapeType],
    tapes: list[TapeType],
    env: Environment[TapeType] | None = None,
    start_tapes: list[TapeType] | None = None,
    reuse_observations: bool = False,
    stop_on_error: bool = False,
) -> int:
    """
    Validate the list of tapes with the agent and environment.
    Check that the agent produce exactly the same steps as the original ones.
    Returns the number of failed tapes.
    """
    ok = 0
    fails = 0
    for i, tape in enumerate(tapes):
        logger.debug(f"Tape {i}")
        try:
            matched = replay_tape(
                agent,
                tape,
                env,
                start_tape=start_tapes[i] if start_tapes else None,
                reuse_observations=reuse_observations,
            )
            if not matched:
                raise FatalError("Tape mismatch")
            ok += 1
        except FatalError as e:
            logger.error(colored(f"Fatal error: {e}, skip tape {i}/{len(tapes)} ({tape.metadata.id})", "red"))
            fails += 1
            if stop_on_error:
                raise e

        logger.debug(colored(f"Ok: {ok}, Fails: {fails}", "green"))
    return fails



import contextlib
import os
import shutil
import typing
from pathlib import Path
from typing import Any, Type

import torch
import transformers
from packaging import version
from transformers import (
    AutoModelForCausalLM,
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    BitsAndBytesConfig,
)
from transformers.models.auto.modeling_auto import _BaseAutoModelClass

from .context import get_accelerator, logger
from .lora import has_lora_checkpoint, lora_load, lora_save, prepare_lora_model
from .types import ModelClass, TrainingMetrics


def is_deepspeed_model(model) -> bool:
    """Check if model is a DeepSpeed engine instance."""
    return model.__class__.__name__.endswith("DeepSpeedEngine")


def get_auto_model_class(
    model_class: ModelClass,
) -> Type[_BaseAutoModelClass]:
    """Get the AutoModel class corresponding to the model class."""
    match model_class:
        case "causal-language-modeling":
            return AutoModelForCausalLM
        case "seq2seq-language-modeling":
            return AutoModelForSeq2SeqLM
        case _:
            raise ValueError(f"Unsupported model class: {model_class}")


def load_tokenizer(config_name):
    tokenizer = AutoTokenizer.from_pretrained(config_name, use_fast=True)
    if not isinstance(tokenizer, transformers.PreTrainedTokenizerFast):
        raise ValueError(f"tokenizer {tokenizer} is not fast")
    if tokenizer.pad_token is None:
        if tokenizer.eos_token is not None:
            tokenizer.pad_token = tokenizer.eos_token
            tokenizer.pad_token_id = tokenizer.eos_token_id
        else:
            tokenizer.add_special_tokens({"pad_token": "[PAD]"})
    if isinstance(tokenizer, transformers.T5TokenizerFast):
        tokenizer.add_special_tokens({"additional_special_tokens": ["<n>", "<t>"]})  # type: ignore
        tokenizer.add_tokens(new_tokens=["▁{", "{", "▁}", "}"])
    return tokenizer


def load_model(args, model_class, current_dir):
    get_accelerator().wait_for_everyone()

    assert not (
        os.path.exists(current_dir / "pytorch_model.bin")
        and os.path.exists(current_dir / "pytorch_model.bin.index.json")
    ), (
        "Found pytorch_model.bin AND pytorch_model.bin.index.json in {current_dir}! "
        "Please remove one of them. "
        "This may happen if combining deepspeed and non-deepsped training"
    )

    model_to_load = args.config_name
    loading_args: dict[str, Any] = dict(
        use_safetensors=args.use_safetensors,
        trust_remote_code=args.trust_remote_code,
        low_cpu_mem_usage=True,  # this is essential for quick model loading as it does not spend time on a random weights initialization. It cuts loading time of a 15B params model from 100 sec to 12 sec.
    )
    if "use_flash_attention" in args.keys() and args.use_flash_attention:
        assert version.parse(transformers.__version__) >= version.parse("4.34.0"), (
            "flash_attention is only supported for transformers>=4.34.0. " "Please upgrade transformers to use it"
        )
        loading_args["use_flash_attention_2"] = args.use_flash_attention

    is_ds_zero_3 = False
    if getattr(get_accelerator().state, "deepspeed_plugin", None):
        del loading_args["low_cpu_mem_usage"]  # deepspeed is not compatible with this option
        is_ds_zero_3 = get_accelerator().state.deepspeed_plugin.zero_stage == 3  # type: ignore

    if args.load_as_bf16:
        loading_args["torch_dtype"] = torch.bfloat16
    if args.lora.enabled:
        if is_ds_zero_3:
            raise Exception("LoRA is not compatible with Deepspeed zero stage 3")
        if args.lora.base_model_8bit:
            loading_args["quantization_config"] = BitsAndBytesConfig(
                load_in_4bit=False,
                load_in_8bit=True,
                llm_int8_has_fp16_weight=args.load_as_bf16,
            )
        elif args.lora.base_model_4bit:
            loading_args["quantization_config"] = BitsAndBytesConfig(
                load_in_4bit=True,
                load_in_8bit=False,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_use_double_quant=False,
                bnb_4bit_compute_dtype=torch.bfloat16,
            )
    if args.auto_device_map:
        loading_args["device_map"] = "auto"
    model_cls = get_auto_model_class(model_class)
    if (
        os.path.exists(current_dir / "pytorch_model.bin")
        or os.path.exists(current_dir / "model.safetensors")
        or os.path.exists(current_dir / "pytorch_model.bin.index.json")
        or os.path.exists(current_dir / "model.safetensors.index.json")
    ):  # resume
        # Size mismatch errors here may be due to improper used of Deepspeed+save_pretrained()
        # instead, always call save_model_only() in all processes

        # when LoRA enabled, always preload the original model, the lora weights will be loaded later
        model_to_load = args.config_name if args.lora.enabled else str(current_dir)
        logger.info(f"Loading model {model_cls} weights from {current_dir}")
    else:  # from scratch
        logger.info(f"Initializing model {model_cls} from {args.config_name}")

    logger.info(f"Loading args: {loading_args}")
    model = model_cls.from_pretrained(model_to_load, **loading_args)

    if args.lora.enabled:
        model = prepare_lora_model(args.lora, model, args.gradient_checkpointing)
        if has_lora_checkpoint(current_dir):
            lora_load(current_dir, model)
    elif args.gradient_checkpointing:
        model.gradient_checkpointing_enable(gradient_checkpointing_kwargs={"use_reentrant": True})

    get_accelerator().wait_for_everyone()
    return model


def save_training_state(
    training_state_dir: Path,
    model: transformers.PreTrainedModel,
    optimizer,
    lr_scheduler,
    extra_training_state: dict[str, typing.Any] = {},
):
    with get_temporary_folder_and_move(training_state_dir) as temp_dir:
        _save_training_state(temp_dir, model, optimizer, lr_scheduler, extra_training_state)


def _save_training_state(
    training_state_dir: Path,
    model: transformers.PreTrainedModel,
    optimizer,
    lr_scheduler,
    extra_training_state: dict[str, typing.Any] = {},
):
    """
    Checkpoint model, optimizer, lr_scheduler, scaler (?) and extra_training_state

    extra_training_state: a dictionary containing completed_steps, passes, best_loss, best_perplexity, etc.

    Compatible with and without deepspeed.
    Make sure to call on *all* accelerate processes.

    Modified from https://github.com/huggingface/accelerate/blob/main/examples/by_feature/deepspeed_with_config_support.py#L247
    More details at: https://deepspeed.readthedocs.io/en/latest/model-checkpointing.html
    """
    assert (
        not os.path.exists(training_state_dir) or training_state_dir.is_dir()
    ), f"output_dir {training_state_dir} must be a directory"

    if is_deepspeed_model(model):
        # Save both model and optimizer, as well as lr_scheduler if supported by deepspeed
        logger.info("Save deepspeed training state")
        client_state = dict(extra_training_state)
        if model.lr_scheduler is None:  # lr_scheduler not handled by deepspeed
            logger.warning(
                f"Manually adding DeepSpeed-unsupported lr_scheduler of type {type(lr_scheduler).__name__} to the checkpoint"
            )
            client_state["lr_scheduler_state"] = lr_scheduler.state_dict()  # manually save state
        success = model.save_checkpoint(training_state_dir, tag="deepspeed", client_state=client_state)
        assert success, f"Failed to save deepspeed training state into {training_state_dir}"
        logger.info(f"Saved deepspeed training state to {training_state_dir}")
    else:  # multi_gpu mode (no deepspeed)
        # Only save training_state in main process
        logger.info("Save accelerate training state")
        if get_accelerator().is_main_process:
            training_state = dict(extra_training_state)
            training_state["optimizer_state"] = optimizer.state_dict()
            training_state["lr_scheduler_state"] = lr_scheduler.state_dict()
            get_accelerator().save(training_state, training_state_dir / "training_state.pt")
            logger.info(f"Saved accelerate training state to {training_state_dir}")


def load_training_checkpoint(
    training_state_dir: Path,
    model: transformers.PreTrainedModel,
    optimizer,
    lr_scheduler,
):
    """
    Load checkpoint created by save_training_checkpoint() in-place:

    - With deepspeed, this will load model, optimizer, lr_scheduler states in-place.
    - Without deepspeed, this will *only* load optimizer, lr_scheduler states in-place,
        but *not* model states!
    """
    assert (
        not os.path.exists(training_state_dir) or training_state_dir.is_dir()
    ), f"output_dir {training_state_dir} must be a directory"

    if is_deepspeed_model(model):
        logger.info("Load deepspeed training state")
        # This magically loads optimizer and lr_scheduler states (if they were saved)
        # (the passed optimizer and lr_scheduler arguments will be ignored)
        load_path, extra_training_state = model.load_checkpoint(
            training_state_dir,
            tag="deepspeed",
            load_optimizer_states=True,
            load_lr_scheduler_states=True,
        )
        if load_path is None:
            raise RuntimeError(f"Loading deepspeed checkpoint from {training_state_dir} failed")
        if (
            model.lr_scheduler is None
            and extra_training_state is not None
            and "lr_scheduler_state" in extra_training_state
        ):
            # Manually load lr_scheduler states
            logger.warning(f"Manually loading ds-unsupported lr_scheduler of type {type(lr_scheduler).__name__}")
            lr_scheduler.load_state_dict(extra_training_state["lr_scheduler_state"])
        logger.info(f"Loaded deepspeed checkpoint from {training_state_dir}")
    else:  # multi_gpu (no deepspeed)
        # This needs to be called from all processes
        training_state = torch.load(training_state_dir / "training_state.pt", map_location="cpu")
        optimizer.load_state_dict(training_state["optimizer_state"])
        lr_scheduler.load_state_dict(training_state["lr_scheduler_state"])
        del training_state["optimizer_state"]
        del training_state["lr_scheduler_state"]
        extra_training_state = training_state
        logger.info(f"Loaded accelerate checkpoint from {training_state_dir}")
    return extra_training_state


@contextlib.contextmanager
def get_temporary_folder_and_move(output_dir: Path):
    """
    Context manager safe checkpointing.

    Creates temporary folder `~output_dir`, then rename to final destination
    """
    if os.path.exists(output_dir) and not os.path.isdir(output_dir):
        raise ValueError("get_temporary_folder_and_move: output_dir is not a directory")

    output_dir = output_dir.resolve()
    temporary_path = output_dir.parent / ("~" + output_dir.name)

    if get_accelerator().is_main_process:
        if os.path.exists(temporary_path):
            logger.info(f"Deleting temporary directory {temporary_path}")
            shutil.rmtree(temporary_path)
        logger.info(f"Creating temporary directory {temporary_path}")
        os.makedirs(temporary_path)

    get_accelerator().wait_for_everyone()
    yield temporary_path
    get_accelerator().wait_for_everyone()

    # Move to final path
    if get_accelerator().is_main_process:
        # delete output_dir if it exists
        if os.path.exists(output_dir):
            logger.info(
                f" -> Deleting {output_dir}. "
                f"If this fails, manually delete it and move {temporary_path} to {output_dir}"
            )
            shutil.rmtree(output_dir)
        logger.info(f" -> Renaming {temporary_path} to {output_dir}")
        os.rename(temporary_path, output_dir)
        logger.info(f"Done moving files to {output_dir}")


def save_model_and_tokenizer(
    output_dir: Path,
    model: transformers.PreTrainedModel,
    tokenizer: transformers.PreTrainedTokenizer | transformers.PreTrainedTokenizerFast,
    lora: bool = False,
    safe_serialization: bool = False,
):
    logger.info("Saving model and tokenizer")
    with get_temporary_folder_and_move(output_dir) as temp_dir:
        save_model_only(
            temp_dir,
            model,
            unwrap=True,
            lora=lora,
            safe_serialization=safe_serialization,
        )
        save_tokenizer_only(temp_dir, tokenizer)


def save_model_only(
    output_dir: Path,
    model,
    unwrap: bool = True,
    lora: bool = False,
    safe_serialization: bool = False,
):
    """
    Save model weights and config.

    Creates the following files in output_dir/ :
        - config.json
    and either:
        - pytorch_model.bin (single-file model), OR
        - pytorch_model-XXXXX-of-XXXXX.bin (multi-file model) and pytorch_model.bin.index.json OR
        - the safetensors versions of the files above

    Note that this does not save optimizer, lr_scheduler, scaler, etc.
    Use only for inference or later JGA evaluation, not for resuming training

    The accelerate version must be called on *all* accelerate processes because all of them must save their shards.
    The DeepSpeed version is only called on the main process because the checkpointing and conversion mechanism will gather the shards from all processes.
    """
    assert not os.path.exists(output_dir) or output_dir.is_dir(), f"output_dir {output_dir} must be a directory"
    get_accelerator().wait_for_everyone()

    logger.info(f"Save model to {output_dir}")

    unwrapped_model = get_accelerator().unwrap_model(model) if unwrap else model
    if lora:
        lora_save(output_dir, unwrapped_model)
        return

    # for non-deepspeed models
    elif isinstance(unwrapped_model, transformers.PreTrainedModel):
        logger.info("Saving model using transformers save_pretrained")
        unwrapped_model.save_pretrained(  # type: ignore
            output_dir,
            is_main_process=get_accelerator().is_main_process,
            save_function=get_accelerator().save,
            state_dict=get_accelerator().get_state_dict(model),
            safe_serialization=safe_serialization,
        )
        logger.info(f"Saved model to {output_dir}")
    else:
        raise ValueError(f"model is neither a deepspeed model nor a transformers.PreTrainedModel: {type(model)}")

    if os.path.exists(output_dir / "model.safetensors") and os.path.exists(output_dir / "model.safetensors.index.json"):
        logger.info("Hide model.safetensors because it utterly confuses the HF model loading code")
        os.rename(output_dir / "model.safetensors", output_dir / "model.safetensors.bak")


def save_tokenizer_only(
    output_dir: Path,
    tokenizer: transformers.PreTrainedTokenizer | transformers.PreTrainedTokenizerFast,
):
    """
    Save only tokenizer to output_dir

    Can be called on *all* processes.
    """
    assert not os.path.exists(output_dir) or output_dir.is_dir(), f"output_dir {output_dir} must be a directory"
    if get_accelerator().is_main_process:
        logger.info(f"Save tokenizer to {output_dir}")
        tokenizer.save_pretrained(output_dir)


def remove_results(current_dir, intermediate_root_dir, training_state_dir, log_dir):
    logger.info("Cleaning up checkpoints and training state")
    if os.path.exists(current_dir):
        shutil.rmtree(current_dir)
    if os.path.exists(log_dir):
        shutil.rmtree(log_dir)
    if os.path.exists(intermediate_root_dir):
        shutil.rmtree(intermediate_root_dir)
    if os.path.exists(training_state_dir):
        if os.path.isdir(training_state_dir):
            shutil.rmtree(training_state_dir)


def load_training_state(
    training_state_dir,
    model,
    optimizer,
    lr_scheduler,
    training_metrics: TrainingMetrics,
):
    get_accelerator().wait_for_everyone()
    training_state = load_training_checkpoint(training_state_dir, model, optimizer, lr_scheduler)
    if training_state is None:
        raise ValueError(f"Could not load training state from {training_state_dir}")

    # Update training_metrics with loaded training state (hasattr check is to avoid potential mismatches between training_metrics and training_state)
    vars(training_metrics).update({key: val for key, val in training_state.items() if hasattr(training_metrics, key)})
    return training_metrics



"""
Entrypoint for streamlit, see https://docs.streamlit.io/
"""

import asyncio

import streamlit as st

from tapeagents.tools.computer import Computer
from tapeagents.tools.computer.steps import (
    KeyPressAction,
    MouseClickAtAction,
    OpenUrlAction,
    RunTerminalCommand,
    TypeTextAction,
)

STREAMLIT_STYLE = """
<style>
    /* Highlight the stop button in red */
    button[kind=header] {
        background-color: rgb(255, 75, 75);
        border: 1px solid rgb(255, 75, 75);
        color: rgb(255, 255, 255);
    }
    button[kind=header]:hover {
        background-color: rgb(255, 51, 51);
    }
     /* Hide the streamlit deploy button */
    .stAppDeployButton {
        visibility: hidden;
    }
    h1 {
        font-size: 2em !important;
    }
</style>
"""


async def main():
    st.markdown(STREAMLIT_STYLE, unsafe_allow_html=True)
    exp_path = ".talk_computer"
    if "computer" not in st.session_state:
        st.session_state.computer = Computer(
            exp_path=exp_path,
            grounding_api_url="https://snow-llmd-grounding-8000.job.console.elementai.com",
        )
    computer = st.session_state.computer
    if "messages" not in st.session_state:
        st.session_state.messages = [{"role": "assistant", "content": "Language controlled computer activated."}]

    for message in st.session_state.messages:
        with st.chat_message(message["role"]):
            st.markdown(message["content"])

    if prompt := st.chat_input():
        with st.chat_message("user"):
            st.write(prompt)
        st.session_state.messages.append({"role": "user", "content": prompt})

        if prompt.startswith("open"):
            url = prompt.split(" ", maxsplit=1)[1]
            if not url.startswith("http"):
                url = "https://" + url
            st.status(f"Opening {url}")
            computer.execute_action(OpenUrlAction(url=url))
            msg = f"Opened {url}"
        elif prompt.startswith("run"):
            command = prompt.split(" ", maxsplit=1)[1]
            st.status(f"Running `{command}`")
            obs = computer.execute_action(RunTerminalCommand(command=command))
            msg = obs.output
            if obs.error:
                msg += [f"\n\nError: {obs.error}"]
        elif prompt.startswith("click"):
            target = prompt.split(" ", maxsplit=1)[1]
            st.status(f"Clicking {target}")
            computer.execute_action(MouseClickAtAction(element_description=target))
            msg = f"Clicked '{target}'"
        elif prompt.startswith("type"):
            text = prompt.split(" ", maxsplit=1)[1]
            st.status(f"Typing '{text}'")
            computer.execute_action(TypeTextAction(text=text))
            msg = "Done"
        elif prompt.startswith("input"):
            text = prompt.split(" ", maxsplit=1)[1]
            st.status(f"Typing '{text}'")
            computer.execute_action(TypeTextAction(text=text))
            computer.execute_action(KeyPressAction(text="Return"))
            msg = "Done"
        elif prompt == "up" or prompt == "scroll up":
            st.status("Scrolling up")
            computer.execute_action(KeyPressAction(text="pageup"))
            msg = "Moved up"
        elif prompt == "down" or prompt == "scroll":
            st.status("Scrolling down")
            computer.execute_action(KeyPressAction(text="pagedown"))
            msg = "Moved down"
        elif prompt.startswith("hotkey"):
            st.status("Pressing hotkey")
            keys = prompt.split(" ", maxsplit=1)[1]
            computer.execute_action(KeyPressAction(text=keys))
            msg = f"Pressed '{keys}'"

        if msg:
            with st.chat_message("assistant"):
                st.write(msg)
            st.session_state.messages.append({"role": "assistant", "content": msg})


if __name__ == "__main__":
    asyncio.run(main())



import logging
import os
import re
from time import sleep
from typing import Any, Callable, Literal
from uuid import uuid4

import gymnasium as gym
import markdownify
import numpy as np
import requests
from browsergym.core.action.highlevel import HighLevelActionSet
from browsergym.core.env import BrowserEnv
from browsergym.utils.obs import (
    IGNORED_AXTREE_PROPERTIES,
    _process_bid,
    flatten_dom_to_str,
    prune_html,
)
from bs4 import BeautifulSoup
from PIL import Image
from pydantic import Field

from tapeagents.core import Action, Observation, StepMetadata
from tapeagents.steps import ImageObservation
from tapeagents.tools.base import StatefulTool
from tapeagents.tools.document_reader import read_document
from tapeagents.tools.grounding import GroundingModel
from tapeagents.tools.simple_browser import PageDownAction, PageObservation, PageUpAction

NODES_WITH_BID = [
    "button",
    "link",
    "combobox",
    "checkbox",
    "textbox",
    "input",
    "radio",
    "menuitem",
    "menuitemcheckbox",
    "menuitemradio",
    "LabelText",
    "tab",
]
logger = logging.getLogger(__name__)


class OpenUrlAction(Action):
    """
    Action that opens a page with the provided URL and returns its first page content.
    Use page_down_action to read subsequent pages.
    """

    kind: Literal["open_url_action"] = "open_url_action"
    url: str = Field(description="URL to navigate to")


class GoForwardAction(Action):
    """
    Action that goes forward in the browser history
    """

    kind: Literal["go_forward_action"] = "go_forward_action"


class GoBackAction(Action):
    """
    Action that goes back in the browser history
    """

    kind: Literal["go_back_action"] = "go_back_action"


class CloseTabAction(Action):
    """
    Action that closes the browser tab
    """

    kind: Literal["close_tab_action"] = "close_tab_action"


class NewTabAction(Action):
    """
    Action that opens a new browser tab
    """

    kind: Literal["new_tab_action"] = "new_tab_action"


class TabFocusAction(Action):
    """
    Action that focuses the tab with the provided index
    """

    kind: Literal["tab_focus_action"] = "tab_focus_action"
    index: int = Field(description="index of the tab to focus")


class PressAction(Action):
    """
    Action that focuses on an element with a given BID and presses a combination of keys.
    Accepts logical key names: Backquote, Minus, Equal, Backslash, Backspace, Tab, Delete, Escape,
    ArrowDown, End, Enter, Home, Insert, PageDown, PageUp, ArrowRight, ArrowUp, F1-F12, Digit0-Digit9, KeyA-KeyZ, etc.
    Supported modifier keys: Shift, Control, Alt, Meta.
    """

    kind: Literal["press_action"] = "press_action"
    bid: str = Field(description="BID of the element to focus")
    key_comb: str = Field(description="key combination to press")


class InputTextAction(Action):
    """
    Action that fills in an input element identified by BID with the provided text
    """

    kind: Literal["input_text_action"] = "input_text_action"
    bid: str = Field(description="BID of the input element to fill")
    text: str = Field(description="text to enter into the element")


class TypeTextAction(Action):
    """
    Action that fills in the currently selected input element.
    Only use after clicking on an input element.
    """

    kind: Literal["type_text_action"] = "type_text_action"
    text: str = Field(description="text to enter into the element")


class HoverAction(Action):
    """
    Action that hovers over an element on the page with the provided BID
    """

    kind: Literal["hover_action"] = "hover_action"
    bid: str = Field(description="BID of the element to hover over")


class MouseHoverAction(Action):
    """
    Action that hovers over an icon or control on the screen
    """

    kind: Literal["mouse_hover_action"] = "mouse_hover_action"
    element_description: str = Field(description="brief description of the element to hover over")


class SelectOptionAction(Action):
    """
    Action that selects an option in a dropdown or combobox element with the provided BID.
    ONLY applicable to dropdown and combobox elements!
    """

    kind: Literal["select_option_action"] = "select_option_action"
    bid: str = Field(description="BID of the dropdown or combobox to select from")
    element_description: str = Field(description="brief description of the dropdown or combobox")
    option: str = Field(description="option to select")


class ClickAction(Action):
    """
    Action that clicks the element on the page with the provided BID
    """

    kind: Literal["click_action"] = "click_action"
    bid: str = Field(description="BID of the element to click")
    button: Literal["left", "middle", "right"] = Field(description="button to click", default="left")
    modifiers: list[Literal["Alt", "Control", "Meta", "Shift"]] = Field(
        description="modifier keys to press", default_factory=list
    )


class MouseClickAction(Action):
    """
    Action that clicks an element on the screen.
    When mentioning a date in the element description, use the format commonly spoken or written by humans,
    such as "2 February 2025," rather than machine-readable formats. The day should come before the month,
    and the year should be written in full (e.g., "3 November 2023" instead of "2023-11-03").
    Only describe one specific element that is currently visible on the screen!
    """

    kind: Literal["mouse_click_action"] = "mouse_click_action"
    element_description: str = Field(description="brief description of the element to click")


class PageScreenshotObservation(ImageObservation):
    kind: Literal["page_screenshot_observation"] = "page_screenshot_observation"


class Browser(StatefulTool):
    """
    Browser tool that can load web pages and interact with their content.
    """

    actions: tuple[type[Action], ...] = (
        ClickAction,
        OpenUrlAction,
        GoBackAction,
        GoForwardAction,
        HoverAction,
        InputTextAction,
        PressAction,
        PageDownAction,
        PageUpAction,
        SelectOptionAction,
    )
    observations: tuple[type[Observation], ...] = (PageObservation, PageScreenshotObservation)
    tab_actions: list[type[Action]] = [CloseTabAction, NewTabAction, TabFocusAction]
    axtree: bool = True
    use_grounding: bool = False
    navigation_only: bool = False
    viewport_chars: int = 32000
    viewport_height: int = 720
    viewport_width: int = 1024
    timeout_ms: int = 30000
    headless: bool = True
    save_video: bool = False
    exp_path: str | None = None
    page_load_time_sec: int = 1
    gym_kwargs: dict = {}
    gym_task: str = "browsergym/openended"
    mock: bool = False

    _env: BrowserEnv = None  # type: ignore
    _current_page: str = ""
    _current_viewport: int = 0
    _n_viewports: int = 1
    _action_map: dict[type[Action], Callable] = {}
    _traces_dir: str | None = None
    _record_video_dir: str | None = None
    _screenshots_dir: str | None = None
    _task_id: str = ""
    _grounding: GroundingModel | None = None
    _last_image: Image.Image | None = None
    _non_browser_doc: bool = False

    def model_post_init(self, __context: Any):
        self._current_page = ""
        self._current_viewport = 1
        self._n_viewports = 1
        if self.use_grounding:
            self._grounding = GroundingModel()
            logger.info("Using grounding model")
            self._action_map = {
                MouseClickAction: self.click_grounded,
                CloseTabAction: self.close_tab,
                TypeTextAction: self.input_text_grounded,
                GoBackAction: self.go_back,
                GoForwardAction: self.go_forward,
                OpenUrlAction: self.goto_page,
                MouseHoverAction: self.hover_grounded,
                NewTabAction: self.new_tab,
                PageDownAction: self.next_page,
                PageUpAction: self.previous_page,
                TabFocusAction: self.tab_focus,
            }
        elif self.navigation_only:
            logger.info("Navigation only mode")
            self._action_map = {
                ClickAction: self.click,
                GoBackAction: self.go_back,
                GoForwardAction: self.go_forward,
                OpenUrlAction: self.goto_page,
                PageDownAction: self.next_page,
                PageUpAction: self.previous_page,
            }
        else:
            self._action_map = {
                ClickAction: self.click,
                SelectOptionAction: self.select_option,
                InputTextAction: self.input_text,
                GoBackAction: self.go_back,
                GoForwardAction: self.go_forward,
                OpenUrlAction: self.goto_page,
                HoverAction: self.hover,
                PageDownAction: self.next_page,
                PageUpAction: self.previous_page,
            }
        self.actions = tuple(self._action_map.keys())
        if self.exp_path:
            assert os.path.isdir(self.exp_path)
            self._traces_dir = os.path.join(self.exp_path, "playwright_traces")
            self._record_video_dir = os.path.join(self.exp_path, "attachments/browser/videos")
            self._screenshots_dir = os.path.join(self.exp_path, "attachments/browser/screenshots")
            os.makedirs(self._traces_dir, exist_ok=True)
            os.makedirs(self._record_video_dir, exist_ok=True)
            os.makedirs(self._screenshots_dir, exist_ok=True)
        if self.mock:
            return
        self._env = gym.make(
            self.gym_task,
            headless=self.headless,
            record_video_dir=self._record_video_dir if self.save_video else None,
            action_mapping=HighLevelActionSet(demo_mode="default", subsets=["coord", "workarena++"]).to_python_code,
            timeout=self.timeout_ms,
            viewport={"width": self.viewport_width, "height": self.viewport_height},
            task_kwargs={"start_url": "about:blank"},
            **self.gym_kwargs,
        )  # type: ignore
        while not isinstance(self._env, BrowserEnv):
            self._env = self._env.env
        self._env.reset()
        self._env.context.tracing.start(screenshots=True, snapshots=True)
        screenshot = self._env.step("noop()")[0]["screenshot"]
        self._save_last_screenshot(screenshot)
        logger.info("Browser initialized")

    def execute_action(self, action: Action) -> PageObservation | PageScreenshotObservation:
        action_type = type(action)
        if action_type in self._action_map:
            return self._action_map[action_type](action)
        raise ValueError(f"Unknown action: {action_type}")

    def start_task(self, task_id: str, seed: int = 1, **kwargs) -> dict:
        self._task_id = task_id
        self._env = gym.make(
            task_id,
            headless=self.headless,
            record_video_dir=self._record_video_dir if self.save_video else None,
            action_mapping=HighLevelActionSet(demo_mode="default").to_python_code,
            timeout=self.timeout_ms,
            **kwargs,
        )  # type: ignore
        start_obs, info = self._env.reset(seed=seed)
        self._env.context.tracing.start(screenshots=True, snapshots=True)
        self._env.chat.add_message(role="assistant", msg="Running TapeAgent...")
        assert self._env.task is not None
        info = {
            "name": self._env.task.get_task_id(),
            "goal": start_obs["goal"],
            "task_info": info["task_info"],
            "video": os.path.basename(self._env.page.video.path()) if self._env.page.video else "",
            "chat_video": os.path.basename(self._env.chat.page.video.path()) if self._env.chat.page.video else "",
        }
        sleep(self.page_load_time_sec)  # wait for the page to load
        return info

    def close(self):
        assert self._traces_dir is not None
        self._env.context.tracing.stop(path=os.path.join(self._traces_dir, f"{self._task_id}.zip"))
        self._env.close()

    def _save_last_screenshot(self, image) -> str:
        if self._screenshots_dir is None:
            return ""
        if isinstance(image, np.ndarray):
            image = Image.fromarray(image)
        if image.mode in ("RGBA", "LA"):
            image = image.convert("RGB")
        if image.size[0] != self.viewport_width or image.size[1] != self.viewport_height:
            image = image.resize((self.viewport_width, self.viewport_height))
        self._last_image = image
        pic_uid = uuid4().hex
        img_path = os.path.join(self._screenshots_dir, f"{pic_uid}.png")
        image.save(img_path)
        return img_path

    def run_browser_action(self, action_text: str) -> PageObservation:
        obs_dict, reward, terminated, truncated, info = self._env.step(action_text)
        error = self.format_error(obs_dict["last_action_error"])
        screen_path = self._save_last_screenshot(obs_dict["screenshot"])
        if self.use_grounding:
            observation = PageScreenshotObservation(image_path=screen_path)
        else:
            if error:
                content = ""
            elif self.axtree:
                content = flatten_axtree(obs_dict["axtree_object"])
            else:
                html_content = prune_html(flatten_dom_to_str(obs_dict["dom_object"]))
                content = self.html_to_markdown(html_content)

            observation = PageObservation(
                text=self.get_viewport(content),
                current_page=self._current_viewport,
                total_pages=self._n_viewports,
                error=error,
                metadata=StepMetadata(other=dict(reward=reward, truncated=truncated, info=info)),
            )
        observation.metadata.other["screenshot_path"] = os.path.relpath(screen_path, self._screenshots_dir)
        observation.metadata.other["env_finished"] = terminated
        return observation

    def html_to_markdown(self, html_content):
        soup = BeautifulSoup(html_content, "html.parser")
        content = markdownify.MarkdownConverter(strip=["img"]).convert_soup(soup)
        content = re.sub(r"\n\s+", "\n", content)
        content = re.sub(r"\n\n+", "\n\n", content)
        return content

    def format_error(self, err: str) -> str:
        logs_separator = "Call log:"
        if logs_separator in err:
            err, logs = err.split(logs_separator)
            logs = "\n".join(logs.split("\n")[:10])
            err = err + f"\n{logs_separator}\n{logs}"
        if not err:
            err = ""
        return err

    def goto_page(self, action: OpenUrlAction) -> PageObservation:
        # if the URL is a local file or a PDF, playwright cannot open it directly, so we read the content
        if action.url.startswith("file://") or action.url.startswith("/"):
            text, error = read_document(action.url)
            self._non_browser_doc = True
            return PageObservation(
                text=self.get_viewport(text),
                current_page=self._current_viewport,
                total_pages=self._n_viewports,
                error=error,
            )
        self._non_browser_doc = False
        obs = self.run_browser_action(f"goto('{action.url}')")
        if obs.error:
            text, error = download_file(action.url)
            obs = PageObservation(
                text=self.get_viewport(text),
                current_page=self._current_viewport,
                total_pages=self._n_viewports,
                error=error,
            )
        return obs

    def click(self, action: ClickAction) -> PageObservation:
        try:
            self.run_browser_action(f"click('{action.bid}'")
        except Exception as e:
            logger.warning(f"Click failed: {e}")
        sleep(self.page_load_time_sec)  # wait for the page to load in case click triggers a page change
        return self.run_browser_action("noop()")

    def click_grounded(self, action: MouseClickAction) -> PageObservation:
        x, y = self._grounding.get_coords(self._last_image, f"click at {action.element_description}")
        logger.info(f"Click at {action.element_description}: {x}, {y}")
        return self.run_browser_action(f"mouse_click({x}, {y})")

    def select_option(self, action: SelectOptionAction) -> PageObservation:
        return self.run_browser_action(f"select_option('{action.bid}', '{action.option}')")

    def hover(self, action: HoverAction) -> PageObservation:
        return self.run_browser_action(f"hover('{action.bid}')")

    def hover_grounded(self, action: MouseHoverAction) -> PageObservation:
        x, y = self._grounding.get_coords(self._last_image, f"click at {action.element_description}")
        logger.info(f"Move cursor at {action.element_description}: {x}, {y}")
        return self.run_browser_action(f"mouse_move({x}, {y})")

    def input_text(self, action: InputTextAction) -> PageObservation:
        text = action.text.replace("'", "\\'")
        return self.run_browser_action(f"fill('{action.bid}', '{text}')")

    def input_text_grounded(self, action: TypeTextAction) -> PageObservation:
        text = action.text.replace("'", "\\'")
        return self.run_browser_action(f"keyboard_type('{text}')")

    def press(self, action: PressAction) -> PageObservation:
        return self.run_browser_action(f"press('{action.bid}', '{action.key_comb}')")

    def tab_focus(self, action: TabFocusAction) -> PageObservation:
        return self.run_browser_action(f"tab_focus({action.index})")

    def new_tab(self, action: NewTabAction) -> PageObservation:
        return self.run_browser_action("new_tab()")

    def close_tab(self, action: CloseTabAction) -> PageObservation:
        return self.run_browser_action("tab_close()")

    def go_back(self, action: GoBackAction) -> PageObservation:
        return self.run_browser_action("go_back()")

    def go_forward(self, action: GoForwardAction) -> PageObservation:
        return self.run_browser_action("go_forward()")

    def next_page(self, action) -> PageObservation:
        if self._current_viewport < self._n_viewports:
            self._current_viewport += 1
            if not self._non_browser_doc:
                self._env.step(f"scroll(0, {self.viewport_height})")
            page = self._current_page[
                self.viewport_chars * (self._current_viewport - 1) : self.viewport_chars * self._current_viewport
            ]
            obs = PageObservation(text=page, current_page=self._current_viewport, total_pages=self._n_viewports)
        else:
            obs = PageObservation(
                text="No more pages to scroll", current_page=self._current_viewport, total_pages=self._n_viewports
            )
        return obs

    def previous_page(self, action) -> PageObservation:
        if self._current_viewport > 1:
            self._current_viewport -= 1
            if not self._non_browser_doc:
                self._env.step(f"scroll(0, -{self.viewport_height})")
            page = self._current_page[
                self.viewport_chars * (self._current_viewport - 1) : self.viewport_chars * self._current_viewport
            ]
            obs = PageObservation(text=page, current_page=self._current_viewport, total_pages=self._n_viewports)
        else:
            obs = PageObservation(
                text="Already at the top page", current_page=self._current_viewport, total_pages=self._n_viewports
            )
        return obs

    def get_viewport(self, content: str) -> str:
        self._current_page = content
        self._n_viewports = len(self._current_page) // self.viewport_chars + 1
        self._current_viewport = 1
        return self._current_page[
            self.viewport_chars * (self._current_viewport - 1) : self.viewport_chars * self._current_viewport
        ]


def flatten_axtree(
    AX_tree,
    extra_properties: dict | None = None,
    with_visible: bool = False,
    with_clickable: bool = False,
    with_center_coords: bool = False,
    with_bounding_box_coords: bool = False,
    with_som: bool = False,
    filter_visible_only: bool = True,
    filter_with_bid_only: bool = False,
    filter_som_only: bool = False,
    coord_decimals: int = 0,
    ignored_properties=IGNORED_AXTREE_PROPERTIES,
    ignore_navigation: bool = False,
    hide_bid_if_invisible: bool = False,
    hide_all_children: bool = False,
    nodes_with_bid: list[str] = NODES_WITH_BID,
) -> str:
    """Formats the accessibility tree into a string text"""
    ignored_roles = ["complementary", "navigation"] if ignore_navigation else []
    extra_properties = extra_properties or {}
    node_id_to_idx = {}
    for idx, node in enumerate(AX_tree["nodes"]):
        node_id_to_idx[node["nodeId"]] = idx

    def dfs(node_idx: int, depth: int, parent_node_filtered: bool) -> str:
        tree_str = ""
        node = AX_tree["nodes"][node_idx]
        indent = "  " * depth
        skip_node = False
        filter_node = False
        node_role = node["role"]["value"]

        if node_role in ignored_roles:
            return tree_str
        elif "name" not in node:
            skip_node = True
            pass
        else:
            node_name = node["name"]["value"]
            if "value" in node and "value" in node["value"]:
                node_value = node["value"]["value"]
            else:
                node_value = None

            attributes = []
            bid = node.get("browsergym_id", None)
            for property in node.get("properties", []):
                if "value" not in property:
                    continue
                if "value" not in property["value"]:
                    continue

                prop_name = property["name"]
                prop_value = property["value"]["value"]

                if prop_name == "browsergym_id":
                    bid = prop_value
                elif prop_name in ignored_properties:
                    continue
                elif prop_name in ("required", "focused", "atomic"):
                    if prop_value:
                        attributes.append(prop_name)
                else:
                    attributes.append(f"{prop_name}={repr(prop_value)}")

            if node_role == "generic" and not attributes:
                skip_node = True
            elif node_role != "StaticText":
                filter_node, extra_attributes_to_print = _process_bid(
                    bid,
                    extra_properties=extra_properties,
                    with_visible=with_visible,
                    with_clickable=with_clickable,
                    with_center_coords=with_center_coords,
                    with_bounding_box_coords=with_bounding_box_coords,
                    with_som=with_som,
                    filter_visible_only=filter_visible_only,
                    filter_with_bid_only=filter_with_bid_only,
                    filter_som_only=filter_som_only,
                    coord_decimals=coord_decimals,
                )

                # if either is True, skip the node
                skip_node = skip_node or filter_node or (hide_all_children and parent_node_filtered)

                # insert extra attributes before regular attributes
                attributes = extra_attributes_to_print + attributes

            # actually print the node string
            if not skip_node:
                if node_role == "paragraph":
                    node_str = ""
                elif node_role == "StaticText":
                    node_str = node_name.strip()
                else:
                    node_repr = node_name.strip()
                    if node_repr and node_role != "checkbox":
                        node_str = f"{node_role} {node_repr}"
                    else:
                        node_str = "-" if node_role == "listitem" else node_role
                    if (
                        not (
                            bid is None
                            or (hide_bid_if_invisible and extra_properties.get(bid, {}).get("visibility", 0) < 0.5)
                        )
                        and node_role in nodes_with_bid
                    ):
                        node_str = f"BID:{bid} " + node_str

                if node_value is not None:
                    node_str += f' value={repr(node["value"]["value"])}'

                if attributes:
                    node_str += ", ".join([""] + attributes)

                if "'Advertisement'" in node_str:
                    return tree_str
                tree_str += f"{indent}{node_str}"

        for child_node_id in node["childIds"]:
            if child_node_id not in node_id_to_idx or child_node_id == node["nodeId"]:
                continue
            # mark this to save some tokens
            child_depth = depth if skip_node else (depth + 1)
            child_str = dfs(node_id_to_idx[child_node_id], child_depth, parent_node_filtered=filter_node or skip_node)
            if child_str and node_role != "link":
                if tree_str:
                    tree_str += "\n"
                tree_str += child_str

        return tree_str

    return dfs(0, 0, False)


def download_file(url: str):
    user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36"
    response = requests.get(url, headers={"User-Agent": user_agent})
    response.raise_for_status()
    return read_document(response)



import os

DB_DEFAULT_FILENAME = "tapedata.sqlite"
ATTACHMENT_DEFAULT_DIR = "attachments"


def common_cache_dir():
    return os.getenv("TAPEAGENTS_CACHE_DIR", ".cache")


def force_cache():
    return bool(os.environ.get("TAPEAGENTS_FORCE_CACHE", 0))


def is_debug_mode():
    return os.environ.get("TAPEAGENTS_DEBUG", None) == "1"


def sqlite_db_path():
    return os.environ.get("TAPEAGENTS_SQLITE_DB", DB_DEFAULT_FILENAME)



import json
import os
import sys
import tempfile
from pathlib import Path
from pprint import pprint

import pytest
import torch

res_path = Path(__file__).parent.resolve() / "res"


def _almost_equal(a, b, eps=1e-2):
    return abs(a - b) < eps


def _check_tuning_artifacts(artifact_paths, deepspeed=False, lora=False):
    artifact_paths_set = set(artifact_paths)
    expected_paths = [
        "current",
        "intermediate",
        "intermediate/16",
        "intermediate/32",
        "log",
        "log/info_0.log",
        "summary.json",
        "training_state",
        "wandb_info.json",
    ]
    if deepspeed:
        expected_paths.append("training_state/deepspeed")
        expected_paths.append("training_state/latest")
    else:
        expected_paths.append("training_state/training_state.pt")

    checkpoint_files = [
        "merges.txt",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
        "vocab.json",
    ]
    if lora:
        checkpoint_files.append("adapter_config.json")
    else:
        checkpoint_files.append("config.json")
        checkpoint_files.append("model.safetensors")
        checkpoint_files.append("generation_config.json")
    for prefix in ["current/", "intermediate/16/", "intermediate/32/"]:
        for checkpoint_file in checkpoint_files:
            expected_paths.append(prefix + checkpoint_file)

    pprint(artifact_paths, stream=sys.stderr)
    for path in expected_paths:
        assert path in artifact_paths_set, f'Path "{path}" not found in artifacts'


def _check_tuning_result(result, golden_result_name, update_golden_result):
    golden_result_filename = res_path / f"finetune/results/{golden_result_name}.json"
    if update_golden_result:
        with open(golden_result_filename, "w") as wf:
            json.dump(result, wf, indent=4, sort_keys=True)
        return

    with open(golden_result_filename) as f:
        golden_result = json.load(f)

    pprint(result, stream=sys.stderr)
    result_keys = set(result.keys())
    assert all(k in result_keys for k in golden_result.keys())
    not_equals = []
    for k, golden in golden_result.items():
        v = result[k]
        same = _almost_equal(v, golden)
        if not same:
            not_equals.append(f"{k} = {v} (expected {golden})")

    assert not len(not_equals), ". ".join(not_equals)


def _run_isolated_tuning(config_name, overrides="", mixed_precision="no", distributed_mode="no"):
    with tempfile.TemporaryDirectory() as tmpdirname:
        _run_tuning(config_name, overrides, tmpdirname, mixed_precision, distributed_mode)
        tmpdirpath = Path(tmpdirname)
        artifacts = _get_run_artifacts(tmpdirpath)
        with open(tmpdirpath / "summary.json") as f:
            result = json.load(f)
    return result, artifacts


def _run_tuning(config_name, overrides, tmpdirname, mixed_precision="no", distributed_mode="no"):
    config_dir = res_path / "conf"
    accelerate_args = _prepare_accelerate_args(mixed_precision, distributed_mode)
    cmd = f"accelerate launch {accelerate_args} {res_path}/../run_finetune.py --config-dir={config_dir} finetune={config_name} {overrides} hydra.run.dir={tmpdirname}"
    print(cmd)
    exitcode = os.system(cmd)
    assert exitcode == 0, f"Tuning failed with code {exitcode}"


def _get_run_artifacts(tmpdirpath):
    artifacts = sorted([str(path).replace(str(tmpdirpath) + "/", "") for path in tmpdirpath.glob("**/*")])

    return artifacts


def _prepare_accelerate_args(mixed_precision, distributed_mode):
    accelerate_args = [f"--mixed_precision={mixed_precision}"]
    if distributed_mode == "multi_gpu":
        accelerate_args += [
            "--config_file conf/accelerate/accelerate_base.yaml",
            "--num_processes 2",
            "--multi_gpu",
        ]
    elif distributed_mode == "deepspeed":
        accelerate_args += [
            "--config_file conf/accelerate/accelerate_base.yaml",
            "--use_deepspeed",
            "--deepspeed_config_file conf/accelerate/deepspeed_stage3_bf16.json",
        ]
    else:
        accelerate_args.append(f"--config_file {res_path}/conf/accelerate/accelerate_local.yaml")
    return " ".join(accelerate_args)


# All the following tests use separate test-time only config test.yaml


@pytest.mark.slow
def test_tuning(update_golden_result=False):
    result, artifacts = _run_isolated_tuning("test")
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "regular", update_golden_result)


@pytest.mark.slow
def test_tuning_grad_checkpoints(update_golden_result=False):
    result, artifacts = _run_isolated_tuning("test", overrides="finetune.gradient_checkpointing=true")
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "gradient_checkpointing", update_golden_result)


@pytest.mark.slow
@pytest.mark.gpu
def test_tuning_bf16(update_golden_result=False):
    result, artifacts = _run_isolated_tuning("test", mixed_precision="bf16")
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "bf16", update_golden_result)


@pytest.mark.slow
@pytest.mark.gpu
def test_tuning_fp16(update_golden_result=False):
    result, artifacts = _run_isolated_tuning("test", mixed_precision="fp16")
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "fp16", update_golden_result)


@pytest.mark.slow
@pytest.mark.gpu
def test_tuning_load_as_bf16(update_golden_result=False):
    result, artifacts = _run_isolated_tuning("test", overrides="finetune.load_as_bf16=true")
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "half_precision", update_golden_result)


@pytest.mark.slow
@pytest.mark.multi_gpu
@pytest.mark.skipif(torch.cuda.device_count() < 2, reason="multi GPU test requires 2+ GPUs")
def test_tuning_multi_gpu(update_golden_result=False):
    assert torch.cuda.device_count() >= 2
    result, artifacts = _run_isolated_tuning("test", distributed_mode="multi_gpu")
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "multi_gpu", update_golden_result)


@pytest.mark.slow
@pytest.mark.gpu
def test_tuning_deepspeed(update_golden_result=False):
    result, artifacts = _run_isolated_tuning("test", distributed_mode="deepspeed")
    _check_tuning_artifacts(artifacts, deepspeed=True)
    _check_tuning_result(result, "deepspeed", update_golden_result)


@pytest.mark.slow
def test_tuning_lora(update_golden_result=False):
    result, artifacts = _run_isolated_tuning(
        "test",
        overrides=(
            "finetune.learning_rate=5e-3 finetune.lora.enabled=true finetune.lora.target_modules=[c_proj,c_attn]"
        ),
    )
    _check_tuning_artifacts(artifacts, lora=True)
    _check_tuning_result(result, "lora", update_golden_result)


@pytest.mark.slow
def test_tuning_resumption(update_golden_result=False):
    with tempfile.TemporaryDirectory() as tmpdirname:
        _run_tuning("test", "finetune.interrupt_train_steps=20", tmpdirname)
        tmpdirpath = Path(tmpdirname)
        with open(tmpdirpath / "summary.json") as f:
            intermediate_result = json.load(f)
        intermediate_artifacts = _get_run_artifacts(tmpdirpath)
        assert "current" in intermediate_artifacts
        assert intermediate_result["completed_steps"] == 20
        # resume training in the same folder
        _run_tuning("test", "", tmpdirname)
        artifacts = _get_run_artifacts(tmpdirpath)
        with open(tmpdirpath / "summary.json") as f:
            result = json.load(f)
        with open(tmpdirpath / "log/info_0.log") as f:
            log_data = f.read()

    assert "Loading model " in log_data
    _check_tuning_artifacts(artifacts)
    _check_tuning_result(result, "resumption", update_golden_result)


if __name__ == "__main__":
    update_golden_result = False
    test_tuning(update_golden_result)
    test_tuning_bf16(update_golden_result)
    test_tuning_fp16(update_golden_result)
    test_tuning_load_as_bf16(update_golden_result)
    test_tuning_grad_checkpoints(update_golden_result)
    test_tuning_deepspeed(update_golden_result)
    test_tuning_lora(update_golden_result)
    test_tuning_resumption(update_golden_result)
    test_tuning_multi_gpu(update_golden_result)



from __future__ import annotations

import warnings
from typing import Any, Literal, Mapping, Sequence, TypeAlias

import jsonref
import jsonschema
import referencing.jsonschema
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    FieldSerializationInfo,
    JsonValue,
    SerializerFunctionWrapHandler,
    TypeAdapter,
    model_serializer,
)
from typing_extensions import Self

from tapeagents.core import Observation

from .error import (
    FormFillerStateError,
    InvalidFunctionParametersError,
    InvalidFunctionParameterSkipError,
    InvalidFunctionParameterValueError,
    InvalidFunctionReturnValueError,
    InvalidFunctionSchemaError,
    UnknownFunctionParameterError,
    UnknownFunctionSchemaError,
)
from .types import FunctionName, ParameterName

JsonType: TypeAlias = Literal["null", "boolean", "object", "array", "number", "string", "integer"]


class JsonSchema(BaseModel):
    """
    This is a simplified version of JSON Schema,
    which is used to describe the schema of a function's parameters and return value.
    See https://json-schema.org/understanding-json-schema/index.html for JSON Schema
    and https://datatracker.ietf.org/doc/html/draft-pbryan-zyp-json-ref-03 for JSON references.
    """

    # camel case (e.g. maxLength) is used in the JSON schema standard
    # some of our data uses underscore-separated names (e.g. max_length)
    # with this setting this class can load both
    model_config = ConfigDict(populate_by_name=True)

    title: None | str = Field(default=None, description="The title of the schema.")
    description: None | str = Field(default=None, description="The description of the schema.")
    type: None | JsonType | list[JsonType] = Field(default=None, description="The type(s) of the schema.")
    format: None | str = Field(default=None, description="The format of the schema.")
    enum: None | list[JsonValue] = Field(default=None, description="The enum values of this schema.")
    const: None | JsonValue = Field(default=None, description="The constant value of this schema.")
    default: None | JsonValue = Field(default=None, description="The default value of this schema.")
    properties: None | dict[str, JsonSchema] = Field(
        default=None,
        description="The properties (key-subschema pairs) of the object that this schema describes.",
    )
    required: None | list[str] = Field(
        default=None,
        description="The list of required properties of the object that this schema describes.",
    )
    items: None | JsonSchema | list[JsonSchema] = Field(
        default=None,
        description="The subschema(s) of the items of the array that this schema describes.",
    )
    all_of: None | list[JsonSchema] = Field(
        alias="allOf",
        default=None,
        description="To validate against allOf, the given data must be valid against all of the given subschemas. This is an intersection type.",
    )
    any_of: None | list[JsonSchema] = Field(
        alias="anyOf",
        default=None,
        description="To validate against anyOf, the given data must be valid against any (one or more) of the given subschemas. This is a union type.",
    )
    one_of: None | list[JsonSchema] = Field(
        alias="oneOf",
        default=None,
        description="To validate against oneOf, the given data must be valid against exactly one of the given subschemas. This is an exclusive union type.",
    )
    minimum: None | int | float = Field(
        default=None,
        description="The minimum value of this schema.",
    )
    maximum: None | int | float = Field(
        default=None,
        description="The maximum value of this schema.",
    )
    exclusive_minimum: None | bool = Field(
        alias="exclusiveMinimum",
        default=None,
        description="Whether the minimum value is exclusive.",
    )
    exclusive_maximum: None | bool = Field(
        alias="exclusiveMaximum",
        default=None,
        description="Whether the maximum value is exclusive.",
    )
    multiple_of: None | int | float = Field(
        alias="multipleOf",
        default=None,
        description="The multiple of this schema.",
    )
    min_length: None | int = Field(
        alias="minLength",
        default=None,
        description="The minimum length of this schema.",
    )
    max_length: None | int = Field(
        alias="maxLength",
        default=None,
        description="The maximum length of this schema.",
    )
    pattern: None | str = Field(
        default=None,
        description="The regular expression pattern of this schema.",
    )
    min_items: None | int = Field(
        alias="minItems",
        default=None,
        description="The minimum number of items of this schema.",
    )
    max_items: None | int = Field(
        alias="maxItems",
        default=None,
        description="The maximum number of items of this schema.",
    )
    unique_items: None | bool = Field(
        alias="uniqueItems",
        default=None,
        description="Whether the items of this schema are unique.",
    )
    prefixItems: None | list[JsonSchema] = Field(
        alias="prefixItems",
        default=None,
        description="The prefix items of this schema.",
    )
    contains: None | JsonSchema = Field(
        default=None,
        description="The schema of the items that this schema contains.",
    )
    minContains: None | int = Field(
        alias="minContains",
        default=None,
        description="The minimum number of items that this schema contains.",
    )
    maxContains: None | int = Field(
        alias="maxContains",
        default=None,
        description="The maximum number of items that this schema contains.",
    )
    additional_properties: None | bool | JsonSchema = Field(
        alias="additionalProperties",
        default=None,
        description="Whether additional properties are allowed for this schema.",
    )
    ref: None | str = Field(alias="$ref", default=None, description="A reference to another schema.")

    @model_serializer(mode="wrap")
    def serialize_model(
        self: Self,
        handler: SerializerFunctionWrapHandler,
        _: FieldSerializationInfo,
    ) -> dict[str, Any]:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            result = handler(self)
        return {k: v for k, v in result.items() if v is not None}

    def get_possible_values(self) -> list[tuple[Any, str]]:
        if self.enum and len(self.enum) > 0 or self.one_of and len(self.one_of) > 0:
            if self.enum:
                possible_values = [(value, str(value)) for value in self.enum if value]
            elif self.one_of:
                possible_values = [(choice.const, str(choice.title)) for choice in self.one_of if choice.const]
            else:
                raise ValueError(f"Unknown schema type: {self}")
        else:
            possible_values = []
        return possible_values


class FunctionSchema(Observation):
    name: FunctionName
    description: str = Field(description="The description of the function.")
    parameters: None | JsonSchema = Field(default=None, description="The JSON schema of the function's parameters.")
    return_value: None | JsonSchema = Field(default=None, description="The JSON schema of the function's return value.")
    definitions: None | dict[str, JsonSchema] = Field(
        default=None,
        description="The definitions (key-subschema pairs) that can be referenced in the parameters and return value schemas.",
    )
    kind: Literal["function_schema"] = "function_schema"

    @model_serializer(mode="wrap")
    def serialize_model(
        self: Self,
        handler: SerializerFunctionWrapHandler,
        _: FieldSerializationInfo,
    ) -> dict[str, Any]:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            result = handler(self)
        return {k: v for k, v in result.items() if v is not None}

    @property
    def with_replaced_refs(self: Self) -> Self:
        """
        Property that returns a copy of the function schema with all references replaced by their target schemas.
        Returns:
            FunctionSchema: The schema with all references replaced by their target schemas.
        """
        # Implementation removed to avoid depenendecy on jsonref
        return self.model_validate(jsonref.replace_refs(self.model_dump(by_alias=True)))

    def get_parameter_schema(self: Self, parameter: ParameterName) -> None | JsonSchema:
        """
        Get the schema of the given parameter.
        Returns `None` if the schema is not available.
        Args:
            parameter (ParameterName): The name of the parameter.
        Returns:
            None | JsonSchema: The schema of the parameter or `None` if not available.
        """
        schema = self
        if schema.parameters is None:
            return None
        if schema.parameters.properties is None:
            return None
        return schema.parameters.properties.get(parameter, None)

    def is_optional_parameter(self: Self, parameter: ParameterName) -> bool:
        """
        Check whether the given parameter is optional.
        Args:
            parameter (ParameterName): The name of the parameter.
        Raises:
            ValueError: If the schema for the parameters is not available.
        Returns:
            bool: Whether the parameter is optional.
        """
        if self.parameters is None:
            raise ValueError(f"Paremeters of function {self.name!r} have no schema.")
        if self.parameters.required is None:
            return False
        return parameter not in self.parameters.required

    def get_parameter_description(self: Self, parameter: ParameterName) -> None | str:
        """
        Get the description of the given parameter.
        Args:
            parameter (ParameterName): The name of the parameter.
        Returns:
            None | str: The description of the parameter or `None` if not available.
        """
        parameter_schema = self.get_parameter_schema(parameter)
        if parameter_schema is None:
            return None
        return parameter_schema.description

    def get_parameter_default(self: Self, parameter: ParameterName) -> None | Any:
        """
        Get the default value of the given parameter.
        Args:
            parameter (ParameterName): The name of the parameter.
        Returns:
            None | Any: The default value of the parameter or `None` if not available.
        """
        parameter_schema = self.get_parameter_schema(parameter)
        if parameter_schema is None:
            return None
        return parameter_schema.default

    @property
    def parameter_names(self: Self) -> list[ParameterName]:
        """
        Get the names of the input parameters.
        Raises:
            ValueError: If the schema for the parameters is not available.
        Returns:
            list[ParameterName]: The names of the input parameters.
        """
        if self.parameters is None:
            raise ValueError(f"Parameters of function {self.name!r} have no schema.")
        if self.parameters.properties is None:
            return []
        return [parameter for parameter in self.parameters.properties.keys()]

    def get_requestable_parameter_names(
        self: Self,
        parameters_filled: Mapping[ParameterName, JsonValue],
        parameters_skipped: Sequence[ParameterName],
    ) -> list[ParameterName]:
        """
        Return a list of parameter names that can be requested.
        The returned list contains all parameter names that are not assigned and not skipped.
        The order of the returned list is the same as the order of the parameters in the schema.
        Args:
            parameters_filled (Mapping[ParameterName, JsonValue]): The parameters that are already filled.
                Only the keys are used.
            parameters_skipped (Sequence[ParameterName]): The parameters that are skipped.
        Returns:
            list[ParameterName]: The names of the parameters that can be requested.
        """
        return [
            parameter
            for parameter in self.parameter_names
            if parameter not in parameters_filled and parameter not in parameters_skipped
        ]

    def get_parameter_enum_values(self: Self, parameter_name: str):
        param_schema = self.get_parameter_schema(parameter_name)
        if param_schema:
            return [value_tuple[0] for value_tuple in param_schema.get_possible_values()]
        return []

    def get_parameter_enum_labels(self: Self, parameter_name: str) -> dict[str, str]:
        param_schema = self.get_parameter_schema(parameter_name)
        if param_schema:
            return {value_tuple[1]: value_tuple[0] for value_tuple in param_schema.get_possible_values()}
        return {}

    def is_parameter_type_enum(self: Self, parameter_name: str):
        param_schema = self.get_parameter_schema(parameter_name)
        if param_schema and param_schema.enum is not None:
            return True
        return False

    def is_parameter_type_one_of_enum(self: Self, parameter_name: str):
        param_schema = self.get_parameter_schema(parameter_name)
        if (
            param_schema
            and param_schema.one_of is not None
            and any(
                one_of_schema.type == "string" and one_of_schema.const is not None
                for one_of_schema in param_schema.one_of
            )
        ):
            return True
        return False

    def validate_parameters_schema(self: Self) -> Literal[True] | FormFillerStateError:
        """
        Validate the parameters JSON schema of the function.
        Returns:
            Literal[True] | FormFillerStateError: `True` if the schema is valid, an error with falsey value otherwise.
        """
        schema = self
        if schema.parameters is None:
            return InvalidFunctionSchemaError(
                function=schema.name,
                function_schema=self,
                message=f"Schema of function {self.name!r} has no parameters.",
            )
        parameters_schema_obj: referencing.jsonschema.Schema = schema.parameters.model_dump(by_alias=True)
        try:
            jsonschema.Draft202012Validator.check_schema(
                schema=parameters_schema_obj,
                format_checker=jsonschema.Draft202012Validator.FORMAT_CHECKER,
            )
        except jsonschema.SchemaError as e:
            return InvalidFunctionSchemaError(
                function=schema.name,
                function_schema=self,
                message=e.message,
            )
        return True

    def validate_parameter_value(
        self: Self,
        parameter: ParameterName,
        parameter_value: JsonValue,
    ) -> Literal[True] | FormFillerStateError:
        """
        Validate the given parameter value against the JSON schema of the given parameter.
        Args:
            parameter (ParameterName): The name of the parameter.
            parameter_value (JsonValue): The value of the parameter.
        Returns:
            Literal[True] | FormFillerStateError: `True` if the value is valid, an error with falsey value otherwise.
        """
        schema = self
        if not isinstance(schema.parameters, JsonSchema):
            return UnknownFunctionParameterError(
                function=self.name,
                parameter=parameter,
                message=f"Schema of function {self.name!r} has no parameters.",
            )
        if schema.parameters.properties is None:
            return UnknownFunctionParameterError(
                function=self.name,
                parameter=parameter,
                message=f"Parameters of function {self.name!r} have no properties.",
            )
        parameter_schema = schema.get_parameter_schema(parameter)
        if parameter_schema is None:
            return UnknownFunctionParameterError(
                function=self.name,
                parameter=parameter,
                message=f"Schema for parameter {parameter!r} is not defined for function {self.name!r}.",
            )
        parameter_schema_obj: referencing.jsonschema.Schema = parameter_schema.model_dump(by_alias=True)
        parameter_value_obj = TypeAdapter(JsonValue).dump_python(parameter_value)
        try:
            jsonschema.validate(
                instance=parameter_value_obj,
                schema=parameter_schema_obj,
                cls=jsonschema.Draft202012Validator,
                format_checker=jsonschema.Draft202012Validator.FORMAT_CHECKER,
            )
        except jsonschema.ValidationError as e:
            return InvalidFunctionParameterValueError(
                function=self.name,
                parameter=parameter,
                parameter_value=parameter_value,
                message=e.message,
            )
        return True

    def validate_parameter_skip(
        self: Self,
        parameter: ParameterName,
    ) -> Literal[True] | FormFillerStateError:
        """
        Validate that the given parameter can be skipped.
        Args:
            parameter (ParameterName): The name of the parameter to skip.
        Returns:
            Literal[True] | FormFillerStateError: `True` if the parameter can be skipped, an error with falsey value otherwise.
        """
        schema = self
        if not isinstance(schema.parameters, JsonSchema):
            return UnknownFunctionParameterError(
                function=self.name,
                parameter=parameter,
                message=f"Schema of function {self.name!r} has no parameters.",
            )
        if schema.parameters.properties is None:
            return UnknownFunctionParameterError(
                function=self.name,
                parameter=parameter,
                message=f"Parameters of function {self.name!r} have no properties.",
            )
        if isinstance(required := schema.parameters.required, list) and parameter in required:
            return InvalidFunctionParameterSkipError(
                function=self.name,
                parameter=parameter,
                message=f"Parameter {parameter!r} of function {self.name!r} is required and cannot be skipped.",
            )
        parameter_schema = schema.get_parameter_schema(parameter)
        if parameter_schema is None:
            return UnknownFunctionParameterError(
                function=self.name,
                parameter=parameter,
                message=f"Schema for parameter {parameter!r} is not defined for function {self.name!r}.",
            )
        if parameter_schema.default is not None:
            return InvalidFunctionParameterSkipError(
                function=self.name,
                parameter=parameter,
                message=f"Parameter {parameter!r} of function {self.name!r} has default value {parameter_schema.default!r} and cannot be skipped.",
            )
        return True

    def validate_parameter_values(
        self: Self,
        parameter_values: dict[ParameterName, JsonValue],
    ) -> Literal[True] | FormFillerStateError:
        """
        Validate the given parameters against the parameters JSON schema of the function.
        Args:
            parameter_values (dict[ParameterName, JsonValue]): The parameters and their values to validate.
        Returns:
            Literal[True] | FormFillerStateError: `True` if the values are valid, an error with falsey value otherwise.
        """
        schema = self
        if not isinstance(schema.parameters, JsonSchema):
            return UnknownFunctionSchemaError(
                function=self.name,
                message=f"Schema of function {self.name!r} has no parameters.",
            )
        parameters_schema_obj: referencing.jsonschema.Schema = schema.parameters.model_dump(by_alias=True)
        parameters_obj = {}
        json_value_adapter = TypeAdapter(JsonValue)
        for parameter, parameter_value in parameter_values.items():
            parameters_obj[parameter] = json_value_adapter.dump_python(parameter_value)
        try:
            jsonschema.validate(
                instance=parameters_obj,
                schema=parameters_schema_obj,
                cls=jsonschema.Draft202012Validator,
                format_checker=jsonschema.Draft202012Validator.FORMAT_CHECKER,
            )
        except jsonschema.ValidationError as e:
            missing_parameters = []
            if e.validator == "required":
                missing_parameters = [
                    parameter for parameter in e.validator_value if parameter not in e.instance.keys()
                ]
            return InvalidFunctionParametersError(
                function=self.name,
                parameters=parameter_values,
                missing_parameters=missing_parameters,
                message=e.message,
            )
        return True

    def validate_return_value_schema(self: Self) -> Literal[True] | FormFillerStateError:
        """
        Validate the return value JSON schema of the function.
        Returns:
            Literal[True] | FormFillerStateError: `True` if the schema is valid, an error with falsey value otherwise.
        """
        schema = self
        if not isinstance(schema.return_value, JsonSchema):
            # It is ok to have no return value schema.
            return True
        return_value_schema_obj: referencing.jsonschema.Schema = schema.return_value.model_dump(by_alias=True)
        try:
            jsonschema.Draft202012Validator.check_schema(
                schema=return_value_schema_obj,
                format_checker=jsonschema.Draft202012Validator.FORMAT_CHECKER,
            )
        except jsonschema.SchemaError as e:
            return InvalidFunctionSchemaError(
                function=schema.name,
                function_schema=self,
                message=e.message,
            )
        return True

    def validate_return_value(
        self: Self,
        return_value: JsonValue,
    ) -> Literal[True] | FormFillerStateError:
        """
        Validate the given return value against the JSON schema of the function.
        Args:
            return_value (JsonValue): The return value.
        Returns:
            Literal[True] | FormFillerStateError: `True` if the value is valid, an error with falsey value otherwise.
        """
        schema = self
        if not isinstance(schema.return_value, JsonSchema):
            # It is ok to have no return value schema.
            return True
        return_value_schema_obj: referencing.jsonschema.Schema = schema.return_value.model_dump(by_alias=True)
        return_value_obj = TypeAdapter(JsonValue).dump_python(return_value)
        try:
            jsonschema.validate(
                instance=return_value_obj,
                schema=return_value_schema_obj,
                cls=jsonschema.Draft202012Validator,
                format_checker=jsonschema.Draft202012Validator.FORMAT_CHECKER,
            )
        except jsonschema.ValidationError as e:
            return InvalidFunctionReturnValueError(
                function=self.name,
                return_value=return_value,
                message=e.message,
            )
        return True



import os
import shutil
from typing import Any, Literal, TypeAlias, Union

from pydantic import Field

from tapeagents.core import (
    LLMOutputParsingFailureAction,
    Observation,
    SetNextNode,
    StopStep,
    Tape,
    TapeMetadata,
    Thought,
)
from tapeagents.dialog_tape import DialogContext, UserStep
from tapeagents.environment import CodeExecutionResult, ExecuteCode
from tapeagents.steps import (
    ActionExecutionFailure,
    ImageObservation,
    ReasoningThought,
    VideoObservation,
    WatchVideoAction,
)
from tapeagents.tool_calling import ToolCallAction, ToolResult
from tapeagents.tools.browser import (
    ClickAction,
    GoBackAction,
    GoForwardAction,
    HoverAction,
    InputTextAction,
    MouseClickAction,
    MouseHoverAction,
    OpenUrlAction,
    PageScreenshotObservation,
    PressAction,
    SelectOptionAction,
    TypeTextAction,
)
from tapeagents.tools.calculator import CalculationResultObservation
from tapeagents.tools.code_executor import PythonCodeAction
from tapeagents.tools.simple_browser import PageDownAction, PageObservation, PageUpAction, ReadDocumentAction
from tapeagents.tools.web_search import SearchAction, SearchResultsObservation


class Plan(Thought):
    """
    Thought that contains the plan to follow when answering the question
    """

    kind: Literal["plan_thought"] = "plan_thought"
    plan: list[str] = Field(description="List of steps to follow when answering the question")


class FactsSurvey(Thought):
    """
    Thought that contains the list of facts needed to answer the question
    """

    kind: Literal["facts_survey_thought"] = "facts_survey_thought"
    given_facts: list[str] = Field(
        description="List of facts already provided in the question",
        default=[],
    )
    facts_to_lookup: list[str] = Field(
        description="List of facts that need to be found on the web or in documents",
        default=[],
    )
    facts_to_derive: list[str] = Field(
        description="List of facts that need to be derived from given facts using reasoning or code execution",
        default=[],
    )
    facts_to_guess: list[str] = Field(
        description="List of facts that need to be guessed from given facts, documents, and reasoning",
        default=[],
    )


class ExtractedFacts(Thought):
    """
    Thought that contains the list of facts extracted from the document
    """

    kind: Literal["extracted_facts_thought"] = "extracted_facts_thought"
    extracted_facts: list[str] | dict[str, Any] | str = Field(description="facts extracted from the observation")


class GaiaQuestion(Observation):
    kind: Literal["question"] = "question"
    content: str
    filename: str | None = None

    @classmethod
    def from_task(cls, question: dict):
        question_prompt = question["Question"]
        filename = None
        if question["file_name"]:
            basename = os.path.basename(question["file_name"])
            tmp_fname = f"/tmp/{basename}"
            shutil.copyfile(question["file_name"], tmp_fname)
            assert os.path.exists(tmp_fname)
            filename = tmp_fname
        return cls(content=question_prompt, filename=filename)


class GaiaAnswer(StopStep):
    """
    Action that indicates the agent has finished the plan and contains the answer or description of failure.
    The answer should use already determined facts without additional conversion!
    Your final answer should be a number OR as few words as possible OR a comma-separated list of numbers and/or strings.
    ADDITIONALLY, your final answer MUST follow any formatting instructions specified in the original question (e.g., alphabetization, sequencing, units, rounding, decimal places, etc.)
    If asked for a number, express it numerically, don't use commas, do not add anything after the number, don't include units such as $ or percent signs unless specified otherwise in the question.
    If asked for a string, don't use articles or abbreviations (e.g. for cities), unless specified otherwise. Don't output any final sentence punctuation such as '.', '!', or '?'.
    If asked for a comma-separated list, apply the above rules depending on whether the elements are numbers or strings.
    If unable to determine the final answer, output an empty string.
    """

    kind: Literal["gaia_answer_action"] = "gaia_answer_action"
    success: bool = Field(description="True if the task was successful, False otherwise")
    overview: str = Field(
        description="List of steps performed to answer the question. If the task was not successful, includes the reason for failure"
    )
    answer_unit: str = Field(description="Unit of measurement for the answer, if applicable; otherwise an empty string")
    answer: Any = Field(description="Short final answer")
    long_answer: str = Field(description="Detailed final answer not restricted by format rules")


THOUGHTS = (ReasoningThought, GaiaAnswer)
GaiaStep: TypeAlias = Union[
    ExtractedFacts,
    ClickAction,
    OpenUrlAction,
    GoBackAction,
    GoForwardAction,
    HoverAction,
    InputTextAction,
    MouseClickAction,
    MouseHoverAction,
    TypeTextAction,
    PressAction,
    PageDownAction,
    PageUpAction,
    SelectOptionAction,
    SearchAction,
    PageDownAction,
    ReadDocumentAction,
    PythonCodeAction,
    ReasoningThought,
    SearchAction,
    WatchVideoAction,
    GaiaAnswer,
    GaiaQuestion,
    SearchResultsObservation,
    PageObservation,
    PageScreenshotObservation,
    ImageObservation,
    VideoObservation,
    CalculationResultObservation,
    CodeExecutionResult,
    ActionExecutionFailure,
    Plan,
    FactsSurvey,
    ExecuteCode,
    LLMOutputParsingFailureAction,
    SetNextNode,
    UserStep,
    ToolCallAction,
    ToolResult,
]


class GaiaMetadata(TapeMetadata):
    task: dict = Field(default_factory=dict)
    result: Any = None
    terminated: bool = False
    attempt_number: int = 0
    level: int = 0
    other: dict = Field(default_factory=dict)


class GaiaTape(Tape[DialogContext, GaiaStep]):
    metadata: GaiaMetadata = Field(default_factory=GaiaMetadata)
    context: DialogContext | None = DialogContext(tools=[])



import json
import logging
import math
import random
import shutil
import traceback
from collections import Counter
from datetime import datetime
from pathlib import Path
from typing import Any, Generator, Literal

import hydra
from hydra.core.global_hydra import GlobalHydra
from hydra.utils import instantiate
from omegaconf import DictConfig
from pydantic import BaseModel
from tqdm import tqdm

from tapeagents.agent import Agent
from tapeagents.core import Error
from tapeagents.dialog_tape import AssistantStep
from tapeagents.io import load_tapes, stream_yaml_tapes
from tapeagents.parallel_processing import lazy_thread_pool_processor

from ..error import UnknownError
from ..tape import FormFillerAgentMetadata, FormFillerContext, FormFillerTape
from ..user_simulator_agent import UserSimulatorAgent, UserSimulatorTape
from .run_formfiller_agent import run_formfiller_agent
from .run_user_simulator import run_user_simulator_agent

logger = logging.getLogger(__name__)


def export_merged_dialogues(output_path: Path, num_layer):
    output_path = Path(output_path)
    logger.info(f"Exporting merged layers from {output_path}")
    all_user_tapes: list[UserSimulatorTape] = []
    all_formfiller_user_tapes: list[FormFillerTape] = []
    all_formfiller_agent_successes: list[FormFillerTape] = []
    all_formfiller_agent_failures: list[FormFillerTape] = []
    for i in range(num_layer):
        logger.info(f"Processing layer {i}")

        if i % 2 == 0:  # user layer
            all_formfiller_user_tapes.extend(
                load_tapes(FormFillerTape, output_path / f"layer_{i}" / "data.yaml")  # type: ignore
            )

            all_user_tapes.extend(
                load_tapes(UserSimulatorTape, output_path / f"layer_{i}" / "user_simulator_tapes.yaml")  # type: ignore
            )
        else:  # agent layer
            all_formfiller_agent_successes.extend(
                load_tapes(FormFillerTape, output_path / f"layer_{i}" / "data.yaml")  # type: ignore
            )

            failures_path = output_path / f"layer_{i}" / "failures.yaml"
            if failures_path.exists():
                all_formfiller_agent_failures.extend(
                    load_tapes(FormFillerTape, failures_path)  # type: ignore
                )

    stats = {
        "formfiller_agent_tapes.yaml": Counter(),
        "formfiller_user_tapes.yaml": Counter(),
        "user_simulator_tapes.yaml": Counter(),
    }
    logger.info(
        f'Saving {len(all_formfiller_agent_successes)} successful agent forks to {output_path / "formfiller_agent_tapes.yaml"}'
    )
    with stream_yaml_tapes(output_path / "formfiller_agent_tapes.yaml") as saver:
        for tape in all_formfiller_agent_successes:
            stats["formfiller_agent_tapes.yaml"][tape.last_action.kind] += 1
            saver.save(tape)

    if all_formfiller_agent_failures:
        stats["formfiller_agent_tape_failures.yaml"] = Counter()
        logger.info(
            f'Saving {len(all_formfiller_agent_failures)} failed agent forks to {output_path / "formfiller_agent_tape_failures.yaml"}'
        )
        with stream_yaml_tapes(output_path / "formfiller_agent_tape_failures.yaml") as saver:
            for tape in all_formfiller_agent_failures:
                stats["formfiller_agent_tape_failures.yaml"][tape.last_action.kind] += 1
                saver.save(tape)

    logger.info(f'Saving {len(all_formfiller_user_tapes)} user forks to {output_path / "formfiller_user_tapes.yaml"}')
    with stream_yaml_tapes(output_path / "formfiller_user_tapes.yaml") as saver:
        for tape in all_formfiller_user_tapes:
            stats["formfiller_user_tapes.yaml"][tape.steps[-1].kind] += 1
            saver.save(tape)

    logger.info(f'Saving {len(all_user_tapes)} user simulator tapes to {output_path / "user_simulator_tapes.yaml"}')
    with stream_yaml_tapes(output_path / "user_simulator_tapes.yaml") as saver:
        for tape in all_user_tapes:
            stats["user_simulator_tapes.yaml"][tape.steps[-1].kind] += 1
            saver.save(tape)

    logger.info(f'Saving stats to {output_path / "stats.json"}')
    with open(output_path / "stats.json", "w") as f:
        json.dump(stats, f, indent=2)


class Layer(BaseModel):
    """
    Layer definitions for user and agent
    """

    who: Literal["user", "agent"]
    count: int
    probs: dict[str, float] = {}


def delete_layer_dialogues(output_path: Path, layer_idx: int):
    # Remove the folder with rmtree
    layer_path = Path(output_path) / f"layer_{layer_idx}"
    if layer_path.exists():
        logger.info(f"-> Deleting layer {layer_idx}")
        shutil.rmtree(layer_path)
        return True
    return False


def resume_last_completed_layer(output_path: Path, force_restart_idx: int = -1) -> tuple[int, list[FormFillerTape]]:
    """
    Return the last completed layer's dialogues and index
    Layers are marked as done by creating a file named 'DONE' in the layer folder.

    If restart_layer_idx >= 0, delete all layers >= force_restart_idx.
    Use restart_layer_idx = -1 to resume from the last finished layer.
    """
    if force_restart_idx >= 0:
        logger.info(f"Deleting all layers >= {force_restart_idx} from {output_path}")
        # Delete all layers
        for layer_idx in range(force_restart_idx, 10000):
            if not delete_layer_dialogues(output_path, layer_idx):
                break

    last_completed_layer_dialogues: list[FormFillerTape] = []
    last_completed_layer_idx = -1
    logger.info(f"Attempting to resume dialogues from {output_path}")
    while True:
        if not (Path(output_path) / f"layer_{last_completed_layer_idx + 1}" / "DONE").exists():
            break
        last_completed_layer_idx += 1
        logger.info(f"-> Layer {last_completed_layer_idx} is done, skipping")
    logger.info(f"-> Last completed layer: {last_completed_layer_idx}")
    if last_completed_layer_idx >= 0:
        # Load dialogues from last completed layer
        last_completed_layer_dialogues = load_tapes(
            FormFillerTape, Path(output_path) / f"layer_{last_completed_layer_idx}" / "data.yaml"
        )
    return last_completed_layer_idx, last_completed_layer_dialogues


def parse_layers(layer_objs: list[dict[str, Any]], global_count: int = -1) -> list[Layer]:
    """
    Parse list of layers from a list of dictionaries.
    For user layers, replace "probs: same" with the probabilities of the previous user layer.
    """
    layers = []
    previous_user_layer = None
    previous_layer = None
    for layer_obj in layer_objs:
        if layer_obj.get("probs") == "same":
            assert layer_obj["who"] == "user", "copy_probs is only supported for user layers"
            assert previous_user_layer is not None, "No previous user layer to copy probabilities from"
            layer_obj["probs"] = previous_user_layer.probs
        if global_count >= 0:
            layer_obj["count"] = global_count
        elif layer_obj.get("count") == "same":
            assert previous_layer is not None, "No previous layer to copy count from"
            layer_obj["count"] = previous_layer.count

        layer = Layer.model_validate(layer_obj)
        layers.append(layer)

        previous_layer = layer
        if layer.who == "user":
            previous_user_layer = layer
    return layers


def make_tape_tree(
    *,
    agent: Agent[FormFillerTape],
    layers: list[Layer],
    user_simulator_agents: dict[str, UserSimulatorAgent],
    env_spec: str,
    previous_layer_idx: int = -1,  # -1 if no previous layer
    previous_layer_dialogues: list[FormFillerTape] = [],  # must be empty at first layer
    num_workers: int = 0,
) -> Generator[tuple[int, list[FormFillerTape], list[FormFillerTape], list[UserSimulatorTape] | None], None, None]:
    if previous_layer_idx == -1:
        assert not previous_layer_dialogues, "previous_layer_dialogues must be empty at the first layer"

        # Generate a single empty dialogue
        previous_layer_dialogues = [
            FormFillerTape(
                metadata=FormFillerAgentMetadata(
                    author="make_tape_tree",
                ),
                context=FormFillerContext(
                    env_spec=env_spec,
                    date=str(datetime.now().date()),
                ),
                steps=[AssistantStep(content="Hi, how can I help you?")],
            )
        ]
    else:
        assert previous_layer_idx >= 0, "previous_layer_idx must be >= 0"

    current_layer_idx = previous_layer_idx + 1

    while current_layer_idx < len(layers):
        assert previous_layer_dialogues, "previous_layer_dialogues must not be empty at previous layers"

        layer = layers[current_layer_idx]
        current_layer_dialogues = []
        current_layer_failures = []

        logger.info(f"*************** Processing Layer {current_layer_idx} ***************")

        if layer.who == "user":
            # (1) get all the tapes that each user behavior can continue
            continuable_tapes_per_behavior = {}
            for behavior in layer.probs:
                user_simulator_agent = user_simulator_agents[behavior]

                # Get valid tapes to continue
                continuable_tapes: list[FormFillerTape] = []
                for tape in previous_layer_dialogues:
                    if user_simulator_agent.can_continue(tape):
                        continuable_tapes.append(tape)
                    else:
                        pass  # logger.debug(f"Skipping tape {tape.metadata.id} as it is not continuable")
                logger.info(
                    f"Continuable tapes for behavior {behavior}: {len(continuable_tapes)}/{len(previous_layer_dialogues)} total tapes"
                )

                random.shuffle(continuable_tapes)  # shuffle in place

                continuable_tapes_per_behavior[behavior] = continuable_tapes
            assert (
                sum(map(len, continuable_tapes_per_behavior.values())) > 0
            ), "No continuable tapes found for any behavior"

            # (2) compute the number of tapes to continue for each behavior based on the config probs
            continuable_counts = {
                behavior: prob for behavior, prob in layer.probs.items() if continuable_tapes_per_behavior[behavior]
            }
            behavior_counts = {
                behavior: math.ceil(layer.count * prob / sum(continuable_counts.values()))
                for behavior, prob in continuable_counts.items()
            }
            logger.info(f"Will continue these behaviors with these counts: {behavior_counts}")
            logger.info(
                f"These behaviors could not be continued: {set(layer.probs.keys()) - set(behavior_counts.keys())}"
            )

            # (3) generate the user simulator tapes
            user_simulator_tapes = []
            for behavior, behavior_count in behavior_counts.items():
                continuable_tapes = continuable_tapes_per_behavior[behavior]

                user_simulator_agent = user_simulator_agents[behavior]

                # Just repeat them to fill up the count
                # TODO: limit the number of UF/AF
                continuable_tapes = continuable_tapes * math.ceil(behavior_count / len(continuable_tapes))
                logger.info(f"Calling run_user_simulator_agent for behavior {behavior} with count {behavior_count}")
                logger.info(f"Continuable tapes after repeating: {len(continuable_tapes)}")

                try:
                    for i, (exception, form_filler_tape, user_simulator_tape) in enumerate(
                        lazy_thread_pool_processor(
                            stream=tqdm(continuable_tapes, desc=f"Making User Layer with behavior {behavior}"),
                            worker_func=lambda tape: run_user_simulator_agent(tape, user_simulator_agent),
                            n_workers=num_workers,
                        )
                    ):
                        if i >= behavior_count:
                            logger.warning(f"Reached limit of behavior_count={behavior_count}")
                            break

                        user_simulator_tapes.append(user_simulator_tape)

                        if exception is not None:
                            logger.error(f"Tape {form_filler_tape.metadata.id} failed with error: {exception}")
                            logger.exception(exception, exc_info=exception)
                            current_layer_failures.append(form_filler_tape)
                        elif any(
                            isinstance(step, Error) for step in form_filler_tape.steps
                        ):  # should not happen since input formfiller_tapes are valid and we only add a UserStep
                            logger.error(f"Tape {form_filler_tape.metadata.id} had LLM parsing failure")
                            current_layer_failures.append(form_filler_tape)
                        else:
                            logger.debug(
                                f"Successfully continued tape {form_filler_tape.metadata.id} with {form_filler_tape.metadata.n_added_steps} steps"
                            )
                            current_layer_dialogues.append(form_filler_tape)

                except Exception:
                    logger.error(f"Skipped behavior {behavior}")
                    traceback.print_exc()

            yield current_layer_idx, current_layer_dialogues, current_layer_failures, user_simulator_tapes

        elif layer.who == "agent":
            for result in lazy_thread_pool_processor(
                stream=tqdm(previous_layer_dialogues, desc="Making Agent Layer"),
                worker_func=lambda tape: run_formfiller_agent(tape, agent),
                n_workers=num_workers,
            ):
                if isinstance(result, Exception):
                    logger.exception(f"Error while running lazy_thread_pool_processor: {result}", exc_info=result)
                    continue
                else:
                    input_tape, result = result
                    if isinstance(result, Exception):
                        logger.exception(f"Error while running run_formfiller_agent: {result}", exc_info=result)
                        predicted_tape = input_tape.model_copy(
                            update=dict(steps=input_tape.steps + [UnknownError(message=str(result))])
                        )
                    else:
                        predicted_tape = result

                if any(isinstance(step, Error) for step in predicted_tape.steps):
                    current_layer_failures.append(predicted_tape)
                    logger.error(f"Agent failed to complete tape {predicted_tape.metadata.id}")
                else:
                    current_layer_dialogues.append(predicted_tape)
                    logger.debug(
                        f"Agent successfully completed tape {predicted_tape.metadata.id} with {predicted_tape.metadata.n_added_steps} steps"
                    )

            yield current_layer_idx, current_layer_dialogues, current_layer_failures, None

        else:
            raise ValueError(f"Unexpected layer.who: {layer.who}")

        previous_layer_dialogues = current_layer_dialogues
        current_layer_idx += 1


def main(cfg: DictConfig):
    # Initialize the agent
    agent = instantiate(cfg.agent)

    if agent.llm.parameters["temperature"] < 0.2:
        raise ValueError("Completion engine temperature must be >= 0.2")

    # Parse all layer specifications
    layers = parse_layers(cfg.layers, cfg.global_count)
    num_workers: int = cfg.num_workers

    # Instantiate all user simulator agents
    logger.info("Instantiating all user simulator agents")
    user_simulator_agents: dict[str, UserSimulatorAgent] = {}
    GlobalHydra.instance().clear()
    hydra.initialize(config_path="../conf", version_base="1.2")
    for layer in layers:
        if layer.who == "user":
            for behavior_name in layer.probs:
                if behavior_name in user_simulator_agents:
                    continue
                my_cfg = hydra.compose(
                    overrides=[
                        f"+user_simulator_agent={behavior_name}",
                        f"+llm@user_simulator_agent.llms={cfg.user_simulator_llm}",
                    ],
                )
                user_simulator_agents[behavior_name] = instantiate(my_cfg.user_simulator_agent)
                if user_simulator_agents[behavior_name].llm.parameters["temperature"] < 0.2:  # type: ignore
                    raise ValueError("User completion engine temperature must be >= 0.2")

    logger.info(f"-> {len(user_simulator_agents)} user simulator agents instantiated")
    logger.info(user_simulator_agents.keys())

    # Reload previous dialogues (layers). Will delete layers if force_restart_idx >= 0
    # previous_layer_idx is the last fully processed layer
    previous_layer_idx, previous_layer_dialogues = resume_last_completed_layer(cfg.output_path, cfg.force_restart_idx)

    for current_layer_idx, current_layer_dialogues, current_layer_failures, user_simulator_tapes in make_tape_tree(
        agent=agent,
        layers=layers,
        user_simulator_agents=user_simulator_agents,
        previous_layer_idx=previous_layer_idx,
        previous_layer_dialogues=previous_layer_dialogues,
        num_workers=num_workers,
        env_spec=cfg.preambles_path,
        # Make sure `preambles_path` matches before_summary.env_spec of the previous layers or risk encountering unexpected behaviors
    ):
        # Save tapes
        logger.info(f"Saving dialogues from layer {current_layer_idx} at {cfg.output_path}")
        save_layer_dialogues(
            cfg.output_path, current_layer_idx, current_layer_dialogues, current_layer_failures, user_simulator_tapes
        )

    export_merged_dialogues(cfg.output_path, num_layer=len(layers))


def save_layer_dialogues(
    output_path: Path,
    layer_idx: int,
    dialogues: list[FormFillerTape],
    failures: list[FormFillerTape],
    user_simulator_tapes: list[UserSimulatorTape] | None = None,
):
    layer_path = Path(output_path) / f"layer_{layer_idx}"
    layer_path.mkdir(parents=True, exist_ok=True)

    if dialogues:
        with stream_yaml_tapes(layer_path / "data.yaml") as saver:
            for dialogue in tqdm(dialogues, desc=f"Saving dialogues for layer {layer_idx}"):
                saver.save(dialogue)

    if failures:
        with stream_yaml_tapes(layer_path / "failures.yaml") as saver:
            for dialogue in tqdm(failures, desc=f"Saving failures for layer {layer_idx}"):
                saver.save(dialogue)

    if user_simulator_tapes:
        with stream_yaml_tapes(layer_path / "user_simulator_tapes.yaml") as saver:
            for tape in tqdm(user_simulator_tapes, desc=f"Saving user simulator tapes for layer {layer_idx}"):
                saver.save(tape)

    with open(layer_path / "DONE", "w") as f:
        f.write("")


@hydra.main(version_base="1.2", config_path="../conf", config_name="make_tape_tree")
def main_wrapper(cfg: DictConfig):
    main(cfg.make_tape_tree)


if __name__ == "__main__":
    main_wrapper()



import json
from typing import Literal, TypeAlias, Union

from pydantic import BaseModel, Field, JsonValue

from tapeagents.core import Action, LLMOutputParsingFailureAction, Observation, SetNextNode, StopStep, Thought
from tapeagents.dialog_tape import AssistantStep, UserStep

from .error import _FormFillerStateError
from .schema import FunctionSchema
from .types import FunctionName, ParameterName


class FunctionCandidate(BaseModel):
    """
    A candidate for a function that was resolved from a user message.
    Attributes:
        function: The name of the function.
        short_description: The short description of the function.
    """

    function: str
    short_description: str


class ExitStep(AssistantStep, StopStep):
    kind: Literal["exit_step"] = "exit_step"

    def llm_view(self) -> str:
        return "Exiting the conversation."


class ResolveFunction(Action):
    query: str | None
    kind: Literal["resolve_function_action"] = "resolve_function_action"

    def llm_view(self) -> str:
        return "Looking for function candidates" + f" for the user query `{self.query}`." if self.query else "."


class FunctionCandidates(Observation):
    candidates: list[FunctionCandidate]
    kind: Literal["function_candidates"] = "function_candidates"

    def llm_view(self) -> str:
        return "These are the most relevant actions I can do:" + "\n".join(
            f"- {candidate.function}: {candidate.short_description}" for candidate in self.candidates
        )


class InspectFunction(Action):
    function: str
    kind: Literal["inspect_function_action"] = "inspect_function_action"

    def llm_view(self) -> str:
        return f"Inspecting the function `{self.function}`."


class RequestFunction(Thought):
    functions: None | list[FunctionName] = Field(
        default=None,
        description="The list of functions to request. If the field is None, the user can select any function.",
    )
    kind: Literal["request_function_thought"] = "request_function_thought"

    def llm_view(self) -> str:
        view = "I should now ask the user to select an action."
        if self.functions:
            view += f"\nAvailable actions: {', '.join(self.functions)}."
        return view


class CallFunction(Action):
    kind: Literal["call_function_action"] = "call_function_action"
    function: FunctionName

    def llm_view(self) -> str:
        return f"Calling the function `{self.function}`."


class FunctionResult(Observation):
    kind: Literal["function_result"] = "function_result"
    function: FunctionName
    result: JsonValue

    def llm_view(self) -> str:
        return f"The results are: {json.dumps(self.result, indent=2)}"


class RequestFunctionParameters(Thought):
    function: FunctionName
    parameters: list[ParameterName]
    kind: Literal["request_function_parameters_thought"] = "request_function_parameters_thought"

    def llm_view(self) -> FunctionName:
        return f"I should ask for {', '.join(self.parameters)}."


class RequestFunctionCallConfirmation(Thought):
    function: FunctionName
    kind: Literal["request_function_call_confirmation_thought"] = "request_function_call_confirmation_thought"

    def llm_view(self) -> str:
        return f"I should now confirm with the user before doing `{self.function}`."


class RequestExitConfirmation(Thought):
    kind: Literal["request_exit_confirmation_thought"] = "request_exit_confirmation_thought"

    def llm_view(self) -> str:
        return "I should ask for confirmation before exiting the conversation."


class UpdateFunctionParameters(Thought):
    function: FunctionName
    assign: dict[ParameterName, JsonValue] = Field(
        default_factory=dict,
        description="The assignments to update parameters with.",
    )
    skip: list[ParameterName] = Field(
        default_factory=list,
        description="The names of parameters to skip.",
    )
    kind: Literal["update_function_parameters_thought"] = "update_function_parameters_thought"

    def llm_view(self) -> str:
        view = []
        if self.assign:
            for p_name, p_value in self.assign.items():
                view.append(f"I note that: {p_name}={str(p_value)}.")
        if self.skip:
            for p_name in self.skip:
                view.append(f"I note that the user skipped {p_name}.")
        return " ".join(view)


class GatherValuesThought(Thought):
    """
    Used in chain of prompting (e.g., multi-node prompter) where each thought is a small step toward generating an agent message.
    This thought is associated with extracting parameters from the user's message.
    """

    kind: Literal["gather_values_thought"] = "gather_values_thought"
    function: FunctionName
    parameters: dict[ParameterName, JsonValue] = Field(
        default_factory=dict,
        description="The extracted parameters.",
    )

    def llm_view(self) -> str:
        view = []
        for p_name, p_value in self.parameters.items():
            view.append(f"(intermediate) I extracted {p_name}={str(p_value)}.")
        return "\n".join(view)


class VerifyValuesThought(Thought):
    """
    Used in chain of prompting (e.g., multi-node prompter) where each thought is a small step toward generating an agent message.
    This thought is associated with verifying parameters, extracted from the user's message.
    """

    kind: Literal["verify_values_thought"] = "verify_values_thought"
    function: FunctionName
    parameters: dict[ParameterName, JsonValue] = Field(
        default_factory=dict,
        description="The verified parameters.",
    )

    def llm_view(self) -> str:
        view = []
        for p_name, value_dict in self.parameters.items():
            view.append(f"(intermediate) I checked {p_name}={str(value_dict['value'])} is {value_dict['status']}")
            if explanation := value_dict.get("explanation", None):
                view[-1] += f" because {explanation}"
            else:
                view[-1] += "."
        return "\n".join(view)


class RefuseInvalidFunctionParameterValue(Thought):
    function: FunctionName
    parameter: ParameterName
    parameter_value: None | JsonValue = Field(
        default=None,
        description="The value the user tried to set for the parameter.",
    )
    kind: Literal["refuse_invalid_function_parameter_value_thought"] = "refuse_invalid_function_parameter_value_thought"

    def llm_view(self) -> str:
        if self.parameter_value:
            return f"I should inform the user that value '{str(self.parameter_value)}' is invalid for parameter {self.parameter}."
        else:
            return f"I should inform the user their provided value is invalid for parameter {self.parameter}."


class RefuseInexistentFunctionParameter(Thought):
    kind: Literal["refuse_inexistent_function_parameter_thought"] = "refuse_inexistent_function_parameter_thought"
    function: FunctionName

    def llm_view(self) -> str:
        return "I should inform the user that this parameter is not available."


class RefuseInvalidFunctionParameterSkip(Thought):
    function: FunctionName
    parameter: ParameterName
    kind: Literal["refuse_invalid_function_parameter_skip_thought"] = "refuse_invalid_function_parameter_skip_thought"

    def llm_view(self) -> str:
        return f"I should inform the user that parameter {self.parameter} cannot be skipped."


class RefuseInexistentFunction(Thought):
    kind: Literal["refuse_inexistent_function_thought"] = "refuse_inexistent_function_thought"

    def llm_view(self) -> str:
        return "I should inform the user that I cannot do what they want."


class RefuseToEngage(Thought):
    kind: Literal["refuse_to_engage_thought"] = "refuse_to_engage_thought"

    def llm_view(self) -> str:
        return "Part of the user message is irrelevant."


class AnswerFromFunctionSchema(Thought):
    function: FunctionName
    answerable_questions: list[str] = Field(
        default=[], description="List of questions derived from the user message that should be answered."
    )
    kind: Literal["answer_from_function_schema_thought"] = "answer_from_function_schema_thought"

    def llm_view(self) -> str:
        return "I should answer the user based on the available documentation."


class NoAnswerFromFunctionSchema(Thought):
    function: FunctionName
    declined_questions: list[str] = Field(
        default=[], description="List of questions derived from the user message that should be declined."
    )
    kind: Literal["no_answer_in_function_schema_thought"] = "no_answer_in_function_schema_thought"

    def llm_view(self) -> str:
        return "I should inform the user that their question cannot be answered based on the available documentation."


class Exit(StopStep):
    kind: Literal["exit_action"] = "exit_action"

    def llm_view(self) -> str:
        return "Exiting the conversation."


FormFillerStep = Union[
    UserStep,
    AssistantStep,
    ResolveFunction,
    FunctionCandidates,
    InspectFunction,
    FunctionSchema,
    RequestFunction,
    CallFunction,
    FunctionResult,
    RequestFunctionParameters,
    RequestFunctionCallConfirmation,
    RequestExitConfirmation,
    UpdateFunctionParameters,
    GatherValuesThought,
    VerifyValuesThought,
    RefuseInvalidFunctionParameterValue,
    RefuseInvalidFunctionParameterSkip,
    RefuseInexistentFunction,
    RefuseInexistentFunctionParameter,
    RefuseToEngage,
    AnswerFromFunctionSchema,
    NoAnswerFromFunctionSchema,
    Exit,
    LLMOutputParsingFailureAction,
    SetNextNode,
    StopStep,
    _FormFillerStateError,
]

I_SHOULD_STEPS = Union[
    RequestFunction,
    RequestFunctionParameters,
    RequestFunctionCallConfirmation,
    RequestExitConfirmation,
    AnswerFromFunctionSchema,
    NoAnswerFromFunctionSchema,
    RefuseInexistentFunction,
    RefuseInexistentFunctionParameter,
    RefuseInvalidFunctionParameterValue,
    RefuseInvalidFunctionParameterSkip,
    RefuseToEngage,
]
I_NOTE_STEPS: TypeAlias = UpdateFunctionParameters
ACTION_STEPS = (AssistantStep, ResolveFunction, InspectFunction, CallFunction, Exit)



import json

from litellm.utils import ChatCompletionMessageToolCall
from pydantic import TypeAdapter

from tapeagents.core import Prompt, Tape
from tapeagents.dialog_tape import (
    AnnotationAction,
    AnnotatorFreeFormThought,
    AssistantStep,
    DialogAnnotator,
    DialogAnnotatorTape,
    DialogContext,
    DialogTape,
    UserStep,
)
from tapeagents.llms import LiteLLM, LLMStream
from tapeagents.renderers import render_dialog_plain_text
from tapeagents.tool_calling import ToolCalls, ToolResult, ToolSpec

_ANNOTATOR_PROMPT: str = """Here is a dialog between a user and an assistant.

{dialog}

You will judge whether everything that the assistant said
IN THEIR LAST MESSAGE is supported by the result of the function calls.

You will output as JSON of the following format:

{{
    "thinking": <your thought process>
    "answer": <output "yes" or "no">
}}

You must output only the JSON and nothing else. Go!
"""


class GroundednessAnnotator(DialogAnnotator):
    """
    Annotates the steps of a given dialog tape based on the results of function calls in the tape.
    Produces the binary correctness of the assistant's last message.
    """

    def make_prompt(self, tape: DialogAnnotatorTape):
        """
        Creates a prompt for the annotator based on the dialog tape.

        Args:
            tape (DialogAnnotatorTape): The dialog tape to be annotated.

        Returns:
            Prompt: The prompt to be used by the annotator.
        """
        return Prompt(
            messages=[
                {"role": "user", "content": _ANNOTATOR_PROMPT.format(dialog=render_dialog_plain_text(tape.context))}
            ]
        )

    def generate_steps(self, tape: Tape, llm_stream: LLMStream):
        """
        Generates annotation steps based on the LLM stream output.

        Args:
            tape: tape with previous steps.
            llm_stream (LLMStream): The stream of text from the language model.

        Yields:
            AnnotatorFreeFormThought: The thought process of the annotator.
            AnnotationAction: The annotation action indicating whether the assistant's response is grounded.

        Raises:
            ValueError: If the answer in the LLM stream output is not "yes" or "no".
        """
        text = llm_stream.get_text()
        result = json.loads(text)
        yield AnnotatorFreeFormThought(content=result["thinking"])
        if result["answer"] == "yes":
            yield AnnotationAction(annotation=dict(grounded=True))
        elif result["answer"] == "no":
            yield AnnotationAction(annotation=dict(grounded=False))
        else:
            raise ValueError("Invalid answer")


TOOL_SCHEMAS = TypeAdapter(list[ToolSpec]).validate_python(
    [
        {
            "type": "function",
            "function": {
                "name": "search_web",
                "description": "Get some information from the web search",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {"type": "string", "description": "The web search query"},
                    },
                    "required": ["query"],
                },
            },
        }
    ]
)


def main():
    dialog1 = DialogTape(
        context=DialogContext(tools=TOOL_SCHEMAS),
        steps=[
            UserStep(content="Tell me about Montreal"),
            ToolCalls(
                tool_calls=[
                    ChatCompletionMessageToolCall(
                        function={"name": "search_web", "arguments": """{"query": "facts about Montreal"}"""}
                    )
                ]
            ),
            ToolResult(content="Montreal is a city in Canada", tool_call_id="1"),
            AssistantStep(content="Montreal is a city in Canada. It is 123 years old."),
        ],
    )

    llm = LiteLLM(model_name="gpt-4o-mini-2024-07-18")
    annotator = GroundednessAnnotator.create(llm)

    # Annotate the dialog
    annotator_tape1 = annotator.annotate(dialog1)
    print(f"Tape:\n{annotator_tape1.model_dump_json(indent=2)}")
    print(f"Annotation:\n{annotator.get_annotation(annotator_tape1)}")

    # Let's replace the last step and re-annotate
    dialog2 = dialog1.model_copy(
        update=dict(steps=dialog1.steps[:-1] + [AssistantStep(content="Montreal is a city in Canada.")])
    )
    annotator_tape2 = annotator.annotate(dialog2)
    print(f"\n\nTape:\n{annotator_tape2.model_dump_json(indent=2)}")
    print(f"Annotation:\n{annotator.get_annotation(annotator_tape2)}")


if __name__ == "__main__":
    main()



import json
import logging
import os
import threading
from typing import Any, Callable

from termcolor import colored

from tapeagents.config import common_cache_dir, force_cache
from tapeagents.utils import FatalError

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

_CACHE_PREFIX = "tool_cache"
_cache = {}


def cached_tool(tool_fn) -> Callable:
    def wrapper(*args, **kwargs):
        fn_name = getattr(tool_fn, "__name__", repr(tool_fn))
        if result := get_from_cache(fn_name, args, kwargs):
            return result
        if force_cache():
            raise FatalError(f"Cache is forced but no cache entry found for {fn_name}({args}, {kwargs})")
        result = tool_fn(*args, **kwargs)
        add_to_cache(fn_name, args, kwargs, result)
        return result

    return wrapper


def get_from_cache(fn_name: str, args: tuple, kwargs: dict) -> Any:
    global _cache
    if not _cache:
        load_cache()
    key = json.dumps((args, kwargs), sort_keys=True)
    result = _cache.get(fn_name, {}).get(key)
    if result is not None:
        logger.info(colored(f"Tool cache hit for {fn_name}", "green"))
    else:
        logger.info(colored(f"Tool cache miss for {fn_name}", "yellow"))
    return result


def load_cache():
    global _cache
    cache_dir = common_cache_dir()
    if os.path.exists(cache_dir):
        for fname in os.listdir(cache_dir):
            if not fname.startswith(_CACHE_PREFIX):
                continue
            with open(os.path.join(cache_dir, fname)) as f:
                for line in f:
                    data = json.loads(line)
                    tool_cache = _cache.get(data["fn_name"], {})
                    key = json.dumps((data["args"], data["kwargs"]), sort_keys=True)
                    tool_cache[key] = data["result"]
                    _cache[data["fn_name"]] = tool_cache
    for k, v in _cache.items():
        logger.info(f"Loaded {len(v)} cache entries for {k}")


def add_to_cache(fn_name: str, args: tuple, kwargs: dict, result: Any):
    global _cache
    logger.info(f"Adding {fn_name} with args {args} and kwargs {kwargs} to cache")
    tool_cache = _cache.get(fn_name, {})
    key = json.dumps((args, kwargs), sort_keys=True)
    tool_cache[key] = result
    _cache[fn_name] = tool_cache
    fname = os.path.join(
        common_cache_dir(), f"{_CACHE_PREFIX}.{fn_name}.{os.getpid()}.{threading.get_native_id()}.jsonl"
    )
    with open(fname, "a") as f:
        f.write(json.dumps({"fn_name": fn_name, "args": args, "kwargs": kwargs, "result": result}) + "\n")



from tapeagents.demo import Demo
from tapeagents.dialog_tape import DialogContext, DialogTape
from tapeagents.llms import LiteLLM
from tapeagents.renderers.basic import BasicRenderer

from .openai_function_calling import TOOL_SCHEMAS, FunctionCallingAgent, MockToolEnvironment


def try_openai_function_calling_interactive_demo():
    llm = LiteLLM(model_name="gpt-3.5-turbo")
    agent = FunctionCallingAgent.create(llm)
    tape = DialogTape(context=DialogContext(tools=TOOL_SCHEMAS), steps=[])
    environment = MockToolEnvironment(llm)
    demo = Demo(agent, tape, environment, BasicRenderer())
    demo.launch()


if __name__ == "__main__":
    try_openai_function_calling_interactive_demo()
